"""
The ``viewer`` provides a way to view the results after the algorithm is run.

===============================================================================
Overview
===============================================================================
The module ``viewer`` provides the mechanism by which to view the results after
the algorithm is run. In order to get the viewer working |PyQt4|_ is required
and a copy of |volumina|_. The ``main`` function actually starts the algorithm
and can be called externally. Configuration files for the viewer are provided
in the examples_ and are entitled viewer.

.. |PyQt4| replace:: ``PyQt4``
.. _PyQt4: http://www.riverbankcomputing.com/software/pyqt
.. |volumina| replace:: ``volumina``
.. _volumina: http://github.com/ilastik/volumina
.. _examples: http://github.com/nanshe-org/nanshe/tree/master/examples

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 09, 2014 08:51:33 EDT$"


# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Copyright 2011-2014, the ilastik developers

"""High-level API.

"""


from nanshe.util import prof

trace_logger = prof.getTraceLogger(__name__)
logger = prof.logging.getLogger(__name__)

import os
import collections
import itertools
import threading

import h5py
import numpy

from PyQt4.QtGui import QApplication
from PyQt4.QtCore import pyqtSignal
from PyQt4.QtCore import QObject

from volumina.multimethods import multimethod
from volumina.pixelpipeline.datasources import SourceABC, RequestABC, ArraySource, ArrayRequest
from volumina.pixelpipeline.datasources import is_pure_slicing
from volumina.layer import GrayscaleLayer, RGBALayer, ColortableLayer, ClickableColortableLayer, AlphaModulatedLayer
from volumina.viewer import Viewer

from nanshe.io import hdf5
from nanshe.util import iters, xnumpy


class HDF5DatasetNotFoundException(Exception):
    """
        An exception raised when a dataset is not found in an HDF5 file.
    """
    pass


@prof.qt_log_class(trace_logger)
class HDF5DataSource(QObject):
    """
        Creates a source that reads from an HDF5 dataset and shapes it in a way
        that Volumina can use.

        Attributes:
              file_handle(h5py.File or str):           A handle for reading the
                                                       HDF5 file or the
                                                       external file path.

              file_path(str):                          External path to the
                                                       file

              dataset_path(str):                       Internal path to the
                                                       dataset

              full_path(str):                          Both external and
                                                       internal paths combined
                                                       as one path

              dataset_shape(tuple of ints):            A tuple representing the
                                                       shape of the dataset in
                                                       each dimension

              dataset_dtype(numpy.dtype or type):      The type of the
                                                       underlying dataset.

              axis_order(tuple of ints):               A tuple representing how
                                                       to reshape the array
                                                       before returning a
                                                       request.
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    def __init__(self, file_handle, internal_path, record_name="", shape=None, dtype=None):
        """
            Constructs an HDF5DataSource using a given file and path to the
            dataset. Optionally, the shape and dtype can be specified.

            Args:
                file_handle(h5py.File or str):          A file handle for the
                                                        HDF5 file.

                internal_path(str):                     path to the dataset
                                                        inside of the HDF5
                                                        file.

                record_name(str):                       which record to extract
                                                        from the HDF5 file.

                shape(tuple of ints):                   shape of underlying
                                                        dataset if not
                                                        specified defaults to
                                                        that of the dataset.

                dtype(numpy.dtype or type):             type of underlying
                                                        dataset if not
                                                        specified defaults to
                                                        that of the dataset.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(HDF5DataSource, self).__init__()

        self.file_handle = None

        self.file_path = ""
        self.dataset_path = ""

        self.full_path = ""

        # Constructed standard shape and dtype if provided
        self.dataset_shape = tuple(shape) if shape is not None else None
        self.dataset_dtype = numpy.dtype(dtype).type if dtype is not None else None

        self.axis_order = [-1, -1, -1, -1, -1]

        # If it is a filename, get the file handle.
        if isinstance(file_handle, str):
            file_handle.rstrip("/")
            file_handle = h5py.File(file_handle)

        self.file_handle = file_handle

        self.file_path = self.file_handle.filename

        internal_path = "/" + internal_path.strip("/")

        self.dataset_path = internal_path

        # Strip name from braces and quotes
        self.record_name = record_name

        self.full_path = self.file_path + self.dataset_path

        # Check to see if the dataset exists in the file.
        # Otherwise throw an exception for it.
        if self.dataset_path not in self.file_handle:
            raise(HDF5DatasetNotFoundException(
                "Could not find the path \"" + self.dataset_path +
                "\" in filename " + "\"" + self.file_path + "\".")
            )

        # Fill in the shape and or dtype information
        # if it doesn't already exist.
        if (self.dataset_shape is None) and (self.dataset_dtype is None):
            dataset = self.file_handle[self.dataset_path]

            if self.record_name:
                self.dataset_shape = dataset.shape + dataset.dtype[self.record_name].shape

                # If the member has a shape
                # then subdtype must be used if not type can be used.
                if dataset.dtype[self.record_name].subdtype is None:
                    self.dataset_dtype = dataset.dtype[self.record_name].type
                else:
                    self.dataset_dtype = dataset.dtype[self.record_name].subdtype[0].type
            else:
                self.dataset_shape = dataset.shape
                self.dataset_dtype = dataset.dtype.type
        elif (self.dataset_shape is None):
            dataset = self.file_handle[self.dataset_path]
            if self.record_name:
                self.dataset_shape = dataset.shape + dataset.dtype[self.record_name].shape
            else:
                self.dataset_shape = dataset.shape
        elif (self.dataset_dtype is None):
            dataset = self.file_handle[self.dataset_path]
            if self.record_name:

                # If the member has a shape then
                # subdtype must be used if not type can be used.
                if dataset.dtype[self.record_name].subdtype is None:
                    self.dataset_dtype = dataset.dtype[self.record_name].type
                else:
                    self.dataset_dtype = dataset.dtype[self.record_name].subdtype[0].type
            else:
                self.dataset_dtype = dataset.dtype.type

        # Using the shape information,
        # determine how to reshape the axes to present the data as we wish.
        if len(self.dataset_shape) == 1:
            self.axis_order = [-1, 0, -1, -1, -1]
        if len(self.dataset_shape) == 2:
            # self.axis_order = [-1, 0, 1, -1, -1]
            self.axis_order = [-1, 1, 0, -1, -1]
        elif (len(self.dataset_shape) == 3) and (self.dataset_shape[2] <= 4):
            # self.axis_order = [-1, 0, 1, -1, 2]
            self.axis_order = [-1, 1, 0, -1, 2]
        elif len(self.dataset_shape) == 3:
            # self.axis_order = [-1, 1, 2, -1, 0]
            self.axis_order = [-1, 2, 1, -1, 0]
        elif len(self.dataset_shape) == 4:
            # self.axis_order = [-1, 1, 2, 3, 0]
            self.axis_order = [-1, 3, 2, 1, 0]
        elif len(self.dataset_shape) == 5:
            # self.axis_order = [0, 1, 2, 3, 4]
            self.axis_order = [0, 3, 2, 1, 4]
        else:
            raise Exception(
                "Unacceptable shape provided for display. " +
                "Found shape to be \"" + str(self.dataset_shape) + "\"."
            )

        # Construct the shape to be singleton if the axis order is irrelevant
        # or the appropriate shape for the reordered axis.
        self.dataset_shape = tuple(
            1 if _ == -1 else self.dataset_shape[_] for _ in self.axis_order
        )

    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    def clean_up(self):
        # Close file
        if self.file_handle is not None:
            self.file_handle.close()
            self.file_handle = None

        self.file_path = None
        self.dataset_path = None
        self.full_path = None
        self.dataset_dtype = None
        self.dataset_shape = None

    def dtype(self):
        return(self.dataset_dtype)

    def shape(self):
        return(self.dataset_shape)

    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('HDF5DataSource: slicing is not pure')

        assert (len(slicing) == len(self.dataset_shape)),\
            "Expect a slicing for a txyzc array."

        slicing = iters.reformat_slices(slicing, self.dataset_shape)

        return(
            HDF5DataRequest(self.file_handle,
                            self.dataset_path,
                            self.axis_order,
                            self.dataset_dtype,
                            slicing,
                            self.record_name)
        )

    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(HDF5DataSource, SourceABC)


@prof.log_class(trace_logger)
class HDF5DataRequest(object):
    """
        Created by an HDF5DataSource to provide a way to request slices of the
        HDF5 file in a nice way.

        Attributes:
            file_handle(h5py.File or str):         A handle for reading the
                                                   HDF5 file or the external
                                                   file path.

            dataset_path(str):                     Internal path to the dataset
            axis_order(tuple of ints):             A tuple representing how to
                                                   reshape the array before
                                                   returning a request.

            dataset_dtype(numpy.dtype or type):    The type of the underlying
                                                   dataset.

            throw_on_not_found(bool):              Whether to throw an
                                                   exception if the dataset is
                                                   not found.

            slicing(tuple of slices):              The slicing request by
                                                   Volumina.
            actual_slicing(tuple of slices):       The actual slicing that will
                                                   be performed on the dataset.

            throw_on_not_found(bool):              Whether to throw an
                                                   exception if the dataset is
                                                   not found.

        Note:
            Before returning the result to Volumina the axes will likely need
            to be transposed. Also, singleton axes will need to be inserted to
            ensure the dimensionality is 5 as Volumina expects. This result
            will be cached inside the request instance. So, if this request
            instance is kept, this won't need to be repeated.
    """

    # TODO: Try to remove throw_on_not_found. This basically would have been thrown earlier. So, we would rather not have this as it is a bit hacky.
    # TODO: Try to remove dataset_dtype as this should be readily available information from the dataset.

    def __init__(self, file_handle, dataset_path, axis_order, dataset_dtype, slicing, record_name="", throw_on_not_found=False):
        """
            Constructs an HDF5DataRequest using a given file and path to the
            dataset. Optionally, throwing can be suppressed if the source is
            not found.

            Args:
                file_handle(h5py.File or str):              A file handle for
                                                            the HDF5 file.

                dataset_path(str):                          Internal path to
                                                            the dataset.

                axis_order(tuple of ints):                  A tuple
                                                            representing how to
                                                            reshape the array
                                                            before returning a
                                                            request.

                dataset_dtype(numpy.dtype or type):         The type of the
                                                            underlying dataset.

                slicing(tuple of ints):                     The slicing to
                                                            extract from the
                                                            HDF5 file.

                record_name(str):                           Name of member to
                                                            retrieve from
                                                            compound type.

                throw_on_not_found(bool):                   Whether to throw an
                                                            exception if the
                                                            dataset is not
                                                            found.
        """

        # TODO: Look at adding assertion check on slices.

        self.file_handle = file_handle
        self.dataset_path = dataset_path
        self.axis_order = axis_order
        self.dataset_dtype = dataset_dtype
        self.record_name = record_name
        self.throw_on_not_found = throw_on_not_found

        self._result = None

        # Clean up slicing. Here self.slicing is the requested slicing.
        # actual_slicing_dict includes a key for each_axis.
        # To construct the list requires a second pass either way.
        self.slicing = list()
        actual_slicing_dict = dict()
        for i, (each_slice, each_axis) in enumerate(iters.izip(slicing, self.axis_order)):
            self.slicing.append(each_slice)
            if each_axis != -1:
                actual_slicing_dict[each_axis] = each_slice

        self.slicing = tuple(self.slicing)

        # As the dictionary sorts by keys,
        # we are ensured to have the slices in the order of the axes.
        self.actual_slicing = actual_slicing_dict.values()

        # Convert to tuple as it is expected.
        self.actual_slicing = tuple(self.actual_slicing)

    def wait(self):
        if self._result is None:
            # Construct a result the size of the slicing
            slicing_shape = iters.len_slices(self.slicing)
            self._result = numpy.zeros(slicing_shape, dtype=self.dataset_dtype)

            try:
                dataset = self.file_handle[self.dataset_path]

                a_result = None
                if self.record_name:
                    # Copy out the bare minimum data.
                    # h5py does not allowing further index on the type
                    # within the compound type.
                    a_result = dataset[
                        self.actual_slicing[:len(dataset.shape)] + (self.record_name,)]
                    # Apply the remaining slicing to the data read.
                    a_result = a_result[len(dataset.shape) * (slice(None),) + self.actual_slicing[len(dataset.shape):]]
                else:
                    a_result = dataset[self.actual_slicing]

                a_result = numpy.array(a_result)

                # Get the axis order without the singleton axes
                the_axis_order = numpy.array(self.axis_order)
                the_axis_order = the_axis_order[the_axis_order != -1]

                # Reorder the axes for Volumina
                a_result = a_result.transpose(the_axis_order)

                # Insert singleton axes to make 5D for Volumina
                for i, each_axis_order in enumerate(self.axis_order):
                    if each_axis_order == -1:
                        a_result = xnumpy.add_singleton_axis_pos(a_result, i)

                self._result[:] = a_result
            except KeyError:
                if self.throw_on_not_found:
                    raise

            logger.debug("Found the result.")

        return self._result

    def getResult(self):
        return self._result

    def cancel(self):
        pass

    def submit(self):
        pass

    # callback( result = result, **kwargs )
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)
assert issubclass(HDF5DataRequest, RequestABC)


@prof.qt_log_class(trace_logger)
class HDF5Viewer(Viewer):
    """
        Extends the Viewer from Volumina so that it provides some additional
        features that are nice for HDF5 sources.

        Attributes:
              file_handle(h5py.File or str):           A handle for reading the
                                                       HDF5 file or the
                                                       external file path.
              file_path(str):                          External path to the
                                                       file

              dataset_path(str):                       Internal path to the
                                                       dataset

              full_path(str):                          Both external and
                                                       internal paths combined
                                                       as one path

              dataset_shape(tuple of ints):            A tuple representing the
                                                       shape of the dataset in
                                                       each dimension

              dataset_dtype(numpy.dtype or type):      The type of the
                                                       underlying dataset.

              axis_order(tuple of ints):               A tuple representing how
                                                       to reshape the array
                                                       before returning a
                                                       request.
    """

    def __init__(self, parent=None):
        super(HDF5Viewer, self).__init__(parent)

    def addGrayscaleHDF5Source(self, source, shape, name=None, direct=False):
        self.dataShape = shape
        layer = GrayscaleLayer(source, direct=direct)
        layer.numberOfChannels = self.dataShape[-1]

        if name:
            layer.name = name
        self.layerstack.append(layer)
        return layer

    def addGrayscaleHDF5Layer(self, a, name=None, direct=False):
        source, self.dataShape = createHDF5DataSource(a, True)
        layer = GrayscaleLayer(source, direct=direct)
        layer.numberOfChannels = self.dataShape[-1]

        if name:
            layer.name = name
        self.layerstack.append(layer)
        return layer

    def addAlphaModulatedHDF5Layer(self, a, name=None):
        source, self.dataShape = createHDF5DataSource(a, True)
        layer = AlphaModulatedLayer(source)
        if name:
            layer.name = name
        self.layerstack.append(layer)
        return layer

    def addRGBAHDF5Layer(self, a, name=None):
        # TODO: Avoid this array indexing as it is a filename.
        assert False
        assert (a.shape[2] >= 3)
        sources = [None, None, None, None]
        for i in range(3):
            sources[i], self.dataShape = createHDF5DataSource(a[..., i], True)
        if(a.shape[3] >= 4):
            sources[3], self.dataShape = createHDF5DataSource(a[..., 3], True)
        layer = RGBALayer(sources[0], sources[1], sources[2], sources[3])
        if name:
            layer.name = name
        self.layerstack.append(layer)
        return layer

    def addRandomColorsHDF5Layer(self, a, name=None, direct=False):
        layer = self.addColorTableLayer(
            a, name, colortable=None, direct=direct
        )
        layer.colortableIsRandom = True
        layer.zeroIsTransparent = True
        return layer

    def addColorTableHDF5Source(self,
                                source,
                                shape,
                                name=None,
                                colortable=None,
                                direct=False,
                                clickFunctor=None):
        self.dataShape = shape

        if colortable is None:
            colortable = self._randomColors()

        layer = None
        if clickFunctor is None:
            layer = ColortableLayer(source, colortable, direct=direct)
        else:
            layer = ClickableColortableLayer(
                self.editor, clickFunctor, source, colortable, direct=direct
            )
        if name:
            layer.name = name

        layer.numberOfChannels = self.dataShape[-1]
        self.layerstack.append(layer)
        return layer

    def addColorTableHDF5Layer(self,
                               a,
                               name=None,
                               colortable=None,
                               direct=False,
                               clickFunctor=None):
        source, self.dataShape = createHDF5DataSource(a, True)

        if colortable is None:
            colortable = self._randomColors()

        layer = None
        if clickFunctor is None:
            layer = ColortableLayer(source, colortable, direct=direct)
        else:
            layer = ClickableColortableLayer(
                self.editor, clickFunctor, source, colortable, direct=direct
            )
        if name:
            layer.name = name

        layer.numberOfChannels = self.dataShape[-1]
        self.layerstack.append(layer)
        return layer




@multimethod(str, bool)
@prof.log_call(trace_logger)
def createHDF5DataSource(full_path, withShape=False):
    # Get a source for the HDF5 file.
    src = HDF5DataSource(full_path)

    if withShape:
        return src, src.shape()
    else:
        return src


@multimethod(str)
@prof.log_call(trace_logger)
def createHDF5DataSource(full_path):
    return createHDF5DataSource(full_path, False)


class HDF5NoFusedSourceException(Exception):
    pass


class HDF5UndefinedShapeDtypeException(Exception):
    pass


@prof.qt_log_class(trace_logger)
class HDF5DataFusedSource(QObject):
    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    def __init__(self, fuse_axis, *data_sources, **kwargs):
        super(HDF5DataFusedSource, self).__init__()

        if len(data_sources) == 0:
            raise HDF5NoFusedSourceException("Have no data sources to fuse.")

        self.fuse_axis = fuse_axis

        self.data_sources = data_sources

        self.data_sources_defined = numpy.array(self.data_sources)
        self.data_sources_defined = self.data_sources_defined[self.data_sources_defined != numpy.array(None)]
        self.data_sources_defined = list(self.data_sources_defined)

        if len(self.data_sources_defined) == 0:
            if ("dtype" in kwargs) and ("shape" in kwargs):
                self.data_dtype = kwargs["dtype"]
                self.data_shape = kwargs["shape"]
            else:
                raise HDF5UndefinedShapeDtypeException(
                    "Have no defined data sources to fuse and " +
                    "no shape or dtype to fallback on."
                )
        else:
            self.data_dtype = self.data_sources_defined[0].dtype()
            self.data_shape = -numpy.ones((5,), dtype=int)
            for each in self.data_sources_defined[1:]:
                try:
                    each_shape = each.shape()
                    each_shape = numpy.array(each_shape)

                    self.data_shape = numpy.array(
                        [self.data_shape, each_shape.astype(int)]).max(axis=0)
                except AttributeError:
                    pass

                each_dtype = each.dtype()
                if not numpy.can_cast(each_dtype, self.data_dtype):
                    if numpy.can_cast(self.data_dtype, each_dtype):
                        self.data_dtype = each_dtype
                    else:
                        raise Exception(
                            "Cannot find safe conversion " +
                            "between self.data_dtype = " +
                            repr(self.data_dtype) + " and each_dtype = " +
                            repr(each_dtype) + "."
                        )

        self.data_shape[self.fuse_axis] = len(self.data_sources)
        self.data_shape = tuple(self.data_shape)

        self.fuse_axis %= len(self.data_shape)
        if self.fuse_axis < 0:
            self.fuse_axis += len(self.data_shape)

    def numberOfChannels(self):
        return self.dataset_shape[-1]

    def clean_up(self):
        self.fuse_axis = None
        self.data_sources = None
        self.data_dtype = None
        self.data_shape = None

    def dtype(self):
        return self.data_dtype

    def shape(self):
        return self.data_shape

    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('HDF5DataFusedSource: slicing is not pure')


        slicing_formatted = []
        slicing_shape = []

        fuse_slicing = None
        non_fuse_slicing = []
        for i, (each_slicing, each_len) in enumerate(iters.izip(slicing, self.data_shape)):
            each_slicing_formatted = None
            if i == self.fuse_axis:
                each_len = len(self.data_sources)
                fuse_slicing = each_slicing_formatted = iters.reformat_slice(
                    each_slicing, each_len
                )
                non_fuse_slicing.append(slice(0, 1, 1))
            else:
                each_slicing_formatted = iters.reformat_slice(
                    each_slicing, each_len
                )
                non_fuse_slicing.append(each_slicing_formatted)

            each_slicing_len = iters.len_slice(
                each_slicing_formatted, each_len
            )

            slicing_formatted.append(each_slicing_formatted)
            slicing_shape.append(each_slicing_len)

        slicing_formatted = tuple(slicing_formatted)
        slicing_shape = tuple(slicing_shape)
        non_fuse_slicing = tuple(non_fuse_slicing)

        selected_data_sources = self.data_sources[fuse_slicing]

        selected_data_requests = []
        for each_data_source in selected_data_sources:
            each_data_request = None
            if each_data_source is not None:
                each_data_request = each_data_source.request(non_fuse_slicing)

            selected_data_requests.append(each_data_request)

        request = HDF5DataFusedRequest(
            self.fuse_axis,
            slicing_shape,
            self.data_dtype,
            *selected_data_requests
        )

        return(request)

    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(HDF5DataFusedSource, SourceABC)


@prof.log_class(trace_logger)
class HDF5DataFusedRequest(object):
    def __init__(self, fuse_axis, data_shape, data_dtype, *data_requests):
        # TODO: Look at adding assertion check on slices.

        self.fuse_axis = fuse_axis
        self.data_shape = data_shape
        self.data_dtype = data_dtype
        self.data_requests = data_requests

        self._result = None

    def wait(self):
        if self._result is None:
            if True:
                self._result = numpy.zeros(
                    self.data_shape, dtype=self.data_dtype
                )

                for i, each_data_request in enumerate(self.data_requests):
                    if each_data_request is not None:
                        each_result = each_data_request.wait()

                        result_view = xnumpy.index_axis_at_pos(
                            self._result, self.fuse_axis, i
                        )
                        each_result_view = xnumpy.index_axis_at_pos(
                            each_result, self.fuse_axis, i
                        )
                        result_view[:] = each_result_view

                logger.debug("Found the result.")

        return self._result

    def getResult(self):
        return self._result

    def cancel(self):
        pass

    def submit(self):
        pass

    # callback( result = result, **kwargs )
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)
assert issubclass(HDF5DataFusedRequest, RequestABC)


class SyncedChannelLayers(object):
    def __init__(self, *layers):
        self.layers = list(layers)
        self.currently_syncing_list = False

        for each_layer in self.layers:
            each_layer.channelChanged.connect(self)


    def __call__(self, channel):
        if not self.currently_syncing_list:
            self.currently_syncing_list = True

            for each_layer in self.layers:
                each_layer.channel = channel

            self.currently_syncing_list = False


class EnumeratedProjectionConstantSource(QObject):
    """
        Creates a source that uses another source (ideally some HDF5 dataset or
        array) and performs a max projection.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_source(a constant SourceABC):      Source to take the max
                                                        projection of.

            axis(int):                                  The axis to take
                                                        compute the max along.

            _shape(tuple of ints):                      The shape of the source
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    @prof.log_call(trace_logger)
    def __init__(self, constant_source, axis=-1):
        """
            Constructs an EnumeratedProjectionConstantSource using a given file
            and path to the dataset. Optionally, the shape and dtype can be
            specified.

            Args:
                constant_source(a constant SourceABC):      Source to take the
                                                            max projection of.

                axis(int):                                  The axis to take
                                                            compute the max
                                                            along.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(EnumeratedProjectionConstantSource, self).__init__()

        self.constant_source = constant_source
        self.axis = axis
        self._shape = self.constant_source.shape()

        self._shape = list(self._shape)
        self._shape[self.axis] = 1
        self._shape = tuple(self._shape)

        # Real hacky solution for caching.
        slicing = (slice(None), ) * len(self._shape)

        self._constant_source_cached = self.constant_source.request(
            slicing
        ).wait()
        self._constant_source_cached = xnumpy.enumerate_masks_max(
            self._constant_source_cached, axis=self.axis
        )

        self._constant_source_cached_array_source = ArraySource(
            self._constant_source_cached
        )
        self._constant_source_cached_array_request = self._constant_source_cached_array_source.request(
            slicing
        )

    @prof.log_call(trace_logger)
    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    @prof.log_call(trace_logger)
    def clean_up(self):
        # Close file
        self.constant_source = None
        self.axis = None

    @prof.log_call(trace_logger)
    def dtype(self):
        return(self.constant_source.dtype())

    @prof.log_call(trace_logger)
    def shape(self):
        return(self._shape)

    @prof.log_call(trace_logger)
    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception(
                'EnumeratedProjectionConstantSource: slicing is not pure'
            )

        # Drop the slicing for the axis where the projection is being applied.
        constant_source_slicing = list(slicing)
        constant_source_slicing[self.axis] = slice(None)
        constant_source_slicing = tuple(constant_source_slicing)

        return(EnumeratedProjectionConstantRequest(
            self._constant_source_cached_array_request,
            self.axis,
            slicing
        ))

    @prof.log_call(trace_logger)
    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    @prof.log_call(trace_logger)
    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    @prof.log_call(trace_logger)
    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(EnumeratedProjectionConstantSource, SourceABC)


class EnumeratedProjectionConstantRequest(object):
    """
        Created by an EnumeratedProjectionConstantSource to provide a way to
        request slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    @prof.log_call(trace_logger)
    def __init__(self, constant_request, axis, slicing):
        """
            Constructs an EnumeratedProjectionConstantRequest using a given
            file and path to the dataset. Optionally, throwing can be
            suppressed if the source is not found.

            Args:
                constant_request(a constant RequestABC):         The request to
                                                                 take the max
                                                                 projection of.

                axis(int):                                       The axis to
                                                                 take the max
                                                                 projection
                                                                 along.

                slicing(tuple of slices):                        Slicing to be
                                                                 returned.
        """

        # TODO: Look at adding assertion check on slices.

        self.constant_request = constant_request
        self.axis = axis
        self.slicing = slicing

        self._result = None

    @prof.log_call(trace_logger)
    def wait(self):
        if self._result is None:
            # Get the result of the request needed
            self._result = self.constant_request.wait()

            # Perform max
            self._result = numpy.max(self._result, axis=self.axis)

            # Add singleton axis where max was performed
            self._result = xnumpy.add_singleton_axis_pos(
                self._result, self.axis
            )

            # Take the slice the viewer wanted
            self._result = self._result[self.slicing]

            logger.debug("Found the result.")

        return self._result

    @prof.log_call(trace_logger)
    def getResult(self):
        return self._result

    @prof.log_call(trace_logger)
    def cancel(self):
        pass

    @prof.log_call(trace_logger)
    def submit(self):
        pass

    # callback( result = result, **kwargs )
    @prof.log_call(trace_logger)
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    @prof.log_call(trace_logger)
    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)

assert issubclass(EnumeratedProjectionConstantRequest, RequestABC)


class ContourProjectionConstantSource(QObject):
    """
        Created by an ContourProjectionConstantSource to provide a way to
        request slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and
    # thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    @prof.log_call(trace_logger)
    def __init__(self, constant_source, axis=-1):
        """
            Constructs an ContourProjectionConstantSource using a given file
            and path to the dataset. Optionally, the shape and dtype.
            can be specified.

            Args:
                constant_source(a constant SourceABC):      Source to take th
                                                            max projection of.

                axis(int):                                  The axis to take
                                                            compute the max
                                                            along.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(ContourProjectionConstantSource, self).__init__()

        self.constant_source = constant_source
        self.axis = axis
        self._shape = self.constant_source.shape()

        # Real hacky solution for caching.
        slicing = (slice(None), ) * len(self._shape)

        self._constant_source_cached = self.constant_source.request(
            slicing
        ).wait()
        for i in iters.irange(self._constant_source_cached.shape[self.axis]):
            _constant_source_cached_i = xnumpy.index_axis_at_pos(
                self._constant_source_cached, self.axis, i
            )
            _constant_source_cached_i[:] = xnumpy.generate_contour(
                _constant_source_cached_i
            )

        self._constant_source_cached_array_source = ArraySource(
            self._constant_source_cached
        )
        self._constant_source_cached_array_request = self._constant_source_cached_array_source.request(
            slicing
        )

    @prof.log_call(trace_logger)
    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    @prof.log_call(trace_logger)
    def clean_up(self):
        # Close file
        self.constant_source = None
        self.axis = None

    @prof.log_call(trace_logger)
    def dtype(self):
        return(self.constant_source.dtype())

    @prof.log_call(trace_logger)
    def shape(self):
        return(self._shape)

    @prof.log_call(trace_logger)
    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception(
                'ContourProjectionConstantSource: slicing is not pure'
            )

        return(ContourProjectionConstantRequest(
            self._constant_source_cached_array_request,
            self.axis,
            slicing
        ))

    @prof.log_call(trace_logger)
    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    @prof.log_call(trace_logger)
    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    @prof.log_call(trace_logger)
    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(ContourProjectionConstantSource, SourceABC)


class ContourProjectionConstantRequest(object):
    """
        Created by an ContourProjectionConstantSource to provide a way to
        request slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    @prof.log_call(trace_logger)
    def __init__(self, constant_request, axis, slicing):
        """
            Constructs an ContourProjectionConstantRequest using a given file
            and path to the dataset. Optionally, throwing can be suppressed if
            the source is not found.

            Args:
                constant_request(a constant RequestABC):         The request to
                                                                 take the max
                                                                 projection of.

                axis(int):                                       The axis to
                                                                 take the max
                                                                 projection
                                                                 along.

                slicing(tuple of slices):                        Slicing to be
                                                                 returned.
        """

        # TODO: Look at adding assertion check on slices.

        self.constant_request = constant_request
        self.axis = axis
        self.slicing = slicing

        self._result = None

    @prof.log_call(trace_logger)
    def wait(self):
        if self._result is None:
            # Get the result of the request needed
            self._result = self.constant_request.wait()

            # Take the slice the viewer wanted
            self._result = self._result[self.slicing]

            logger.debug("Found the result.")

        return self._result

    @prof.log_call(trace_logger)
    def getResult(self):
        return self._result

    @prof.log_call(trace_logger)
    def cancel(self):
        pass

    @prof.log_call(trace_logger)
    def submit(self):
        pass

    # callback( result = result, **kwargs )
    @prof.log_call(trace_logger)
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    @prof.log_call(trace_logger)
    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)

assert issubclass(ContourProjectionConstantRequest, RequestABC)


class FloatProjectionConstantSource(QObject):
    """
        Created by an FloatProjectionConstantSource to provide a way to request
        slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    @prof.log_call(trace_logger)
    def __init__(self, constant_source):
        """
            Constructs an FloatProjectionConstantSource using a given file and
            path to the dataset. Optionally, the shape and dtype can be
            specific.

            Args:
                constant_source(a constant SourceABC):       Source to make
                                                             float.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(FloatProjectionConstantSource, self).__init__()

        self.constant_source = constant_source
        self._shape = self.constant_source.shape()

    @prof.log_call(trace_logger)
    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    @prof.log_call(trace_logger)
    def clean_up(self):
        # Close file
        self.constant_source = None
        self.axis = None

    @prof.log_call(trace_logger)
    def dtype(self):
        return(numpy.float64)

    @prof.log_call(trace_logger)
    def shape(self):
        return(self._shape)

    @prof.log_call(trace_logger)
    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception(
                'FloatProjectionConstantSource: slicing is not pure'
            )

        return(FloatProjectionConstantRequest(
            self.constant_source.request(slicing)
        ))

    @prof.log_call(trace_logger)
    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    @prof.log_call(trace_logger)
    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    @prof.log_call(trace_logger)
    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(FloatProjectionConstantSource, SourceABC)


class FloatProjectionConstantRequest(object):
    """
        Created by an FloatProjectionConstantSource to provide a way to request
        slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    @prof.log_call(trace_logger)
    def __init__(self, constant_request):
        """
            Constructs an FloatProjectionConstantRequest using a given file and
            path to the dataset. Optionally, throwing can be suppressed if the
            source is not found.

            Args:
                constant_request(a constant RequestABC):         The request to
                                                                 take the max
                                                                 projection of.

                axis(int):                                       The axis to
                                                                 take the max
                                                                 projection
                                                                 along.

                slicing(tuple of slices):                        Slicing to be
                                                                 returned.
        """

        self.constant_request = constant_request

        self._result = None

    @prof.log_call(trace_logger)
    def wait(self):
        if self._result is None:
            # Get the result of the request needed
            self._result = self.constant_request.wait()

            # Convert to float
            self._result = self._result.astype(numpy.float)

            logger.debug("Found the result.")

        return self._result

    @prof.log_call(trace_logger)
    def getResult(self):
        return self._result

    @prof.log_call(trace_logger)
    def cancel(self):
        pass

    @prof.log_call(trace_logger)
    def submit(self):
        pass

    # callback( result = result, **kwargs )
    @prof.log_call(trace_logger)
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    @prof.log_call(trace_logger)
    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)

assert issubclass(FloatProjectionConstantRequest, RequestABC)


class MaxProjectionConstantSource(QObject):
    """
        Creates a source that uses another source (ideally some HDF5 dataset or
        array) and performs a max projection.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_source(a constant SourceABC):      Source to take the max
                                                        projection of.

            axis(int):                                  The axis to take
                                                        compute the max along.

            _shape(tuple of ints):                      The shape of the
                                                        source.
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    @prof.log_call(trace_logger)
    def __init__(self, constant_source, axis=-1):
        """
            Constructs an MaxProjectionConstantSource using a given file and
            path to the dataset. Optionally, the shape and dtype can be
            specified.

            Args:
                constant_source(a constant SourceABC):      Source to take the
                                                            max projection of.

                axis(int):                                  The axis to take
                                                            compute the max
                                                            along.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(MaxProjectionConstantSource, self).__init__()

        self.constant_source = constant_source
        self.axis = axis
        self._shape = self.constant_source.shape()

        self._shape = list(self._shape)
        self._shape[self.axis] = 1
        self._shape = tuple(self._shape)

        # Real hacky solution for caching.
        slicing = (slice(None), ) * len(self._shape)

        self._constant_source_cached = self.constant_source.request(
            slicing
        ).wait()
        self._constant_source_cached = self._constant_source_cached.max(
            axis=self.axis
        )
        self._constant_source_cached = xnumpy.add_singleton_axis_pos(
            self._constant_source_cached, self.axis
        )

        self._constant_source_cached_array_source = ArraySource(
            self._constant_source_cached
        )
        self._constant_source_cached_array_request = self._constant_source_cached_array_source.request(
            slicing
        )

    @prof.log_call(trace_logger)
    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    @prof.log_call(trace_logger)
    def clean_up(self):
        # Close file
        self.constant_source = None
        self.axis = None

    @prof.log_call(trace_logger)
    def dtype(self):
        return(self.constant_source.dtype())

    @prof.log_call(trace_logger)
    def shape(self):
        return(self._shape)

    @prof.log_call(trace_logger)
    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('MaxProjectionConstantSource: slicing is not pure')

        # Drop the slicing for the axis where the projection is being applied.
        constant_source_slicing = list(slicing)
        constant_source_slicing[self.axis] = slice(None)
        constant_source_slicing = tuple(constant_source_slicing)

        return(MaxProjectionConstantRequest(
            self._constant_source_cached_array_request,
            self.axis,
            slicing
        ))

    @prof.log_call(trace_logger)
    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    @prof.log_call(trace_logger)
    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    @prof.log_call(trace_logger)
    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(MaxProjectionConstantSource, SourceABC)


class MaxProjectionConstantRequest(object):
    """
        Created by an MaxProjectionConstantSource to provide a way to request
        slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the max
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the max projection
                                                             along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    @prof.log_call(trace_logger)
    def __init__(self, constant_request, axis, slicing):
        """
            Constructs an MaxProjectionConstantRequest using a given file and
            path to the dataset. Optionally, throwing can be suppressed if the
            source is not found.

            Args:
                constant_request(a constant RequestABC):         The request to
                                                                 take the max
                                                                 projection of.

                axis(int):                                       The axis to
                                                                 take the max
                                                                 projection
                                                                 along.

                slicing(tuple of slices):                        Slicing to be
                                                                 returned.
        """

        # TODO: Look at adding assertion check on slices.

        self.constant_request = constant_request
        self.axis = axis
        self.slicing = slicing

        self._result = None

    @prof.log_call(trace_logger)
    def wait(self):
        if self._result is None:
            # Get the result of the request needed
            self._result = self.constant_request.wait()

            # Perform max
            self._result = numpy.max(self._result, axis=self.axis)

            # Add singleton axis where max was performed
            self._result = xnumpy.add_singleton_axis_pos(
                self._result, self.axis
            )

            # Take the slice the viewer wanted
            self._result = self._result[self.slicing]

            logger.debug("Found the result.")

        return self._result

    @prof.log_call(trace_logger)
    def getResult(self):
        return self._result

    @prof.log_call(trace_logger)
    def cancel(self):
        pass

    @prof.log_call(trace_logger)
    def submit(self):
        pass

    # callback( result = result, **kwargs )
    @prof.log_call(trace_logger)
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    @prof.log_call(trace_logger)
    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)

assert issubclass(MaxProjectionConstantRequest, RequestABC)


class MeanProjectionConstantSource(QObject):
    """
        Creates a source that uses another source (ideally some HDF5 dataset or
        array) and performs a mean projection.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_source(a constant SourceABC):      Source to take the mean
                                                        projection of.

            axis(int):                                  The axis to take
                                                        compute the mean along.

            _shape(tuple of ints):                      The shape of the
                                                        source.
    """

    # TODO: Reshaping should probably be some sort of lazyflow operator and
    # thus removed from this directly.

    isDirty = pyqtSignal(object)
    numberOfChannelsChanged = pyqtSignal(int) # Never emitted

    @prof.log_call(trace_logger)
    def __init__(self, constant_source, axis=-1):
        """
            Constructs an MeanProjectionConstantSource using a given file and
            path to the dataset. Optionally, the shape and dtype can be
            specified.

            Args:
                constant_source(a constant SourceABC):      Source to take the
                                                            mean projection of.

                axis(int):                                  The axis to take
                                                            compute the mean
                                                            along.
        """
        # TODO: Get rid of shape and dtype as arguments.

        super(MeanProjectionConstantSource, self).__init__()

        self.constant_source = constant_source
        self.axis = axis
        self._shape = self.constant_source.shape()

        self._shape = list(self._shape)
        self._shape[self.axis] = 1
        self._shape = tuple(self._shape)

        # Real hacky solution for caching.
        slicing = (slice(None), ) * len(self._shape)

        self._constant_source_cached = self.constant_source.request(
            slicing
        ).wait()
        self._constant_source_cached = self._constant_source_cached.mean(
            axis=self.axis
        )
        self._constant_source_cached = xnumpy.add_singleton_axis_pos(
            self._constant_source_cached, self.axis
        )

        self._constant_source_cached_array_source = ArraySource(
            self._constant_source_cached
        )
        self._constant_source_cached_array_request = self._constant_source_cached_array_source.request(
            slicing
        )

    @prof.log_call(trace_logger)
    def numberOfChannels(self):
        return(self.dataset_shape[-1])

    @prof.log_call(trace_logger)
    def clean_up(self):
        # Close file
        self.constant_source = None
        self.axis = None

    @prof.log_call(trace_logger)
    def dtype(self):
        return(self.constant_source.dtype())

    @prof.log_call(trace_logger)
    def shape(self):
        return(self._shape)

    @prof.log_call(trace_logger)
    def request(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception(
                'MeanProjectionConstantSource: slicing is not pure'
            )

        # Drop the slicing for the axis where the projection is being applied.
        constant_source_slicing = list(slicing)
        constant_source_slicing[self.axis] = slice(None)
        constant_source_slicing = tuple(constant_source_slicing)

        return(MeanProjectionConstantRequest(
            self._constant_source_cached_array_request,
            self.axis,
            slicing
        ))

    @prof.log_call(trace_logger)
    def setDirty(self, slicing):
        if not is_pure_slicing(slicing):
            raise Exception('dirty region: slicing is not pure')
        self.isDirty.emit(slicing)

    @prof.log_call(trace_logger)
    def __eq__(self, other):
        if other is None:
            return False

        return(self.full_path == other.full_path)

    @prof.log_call(trace_logger)
    def __ne__(self, other):
        if other is None:
            return True

        return(self.full_path != other.full_path)

assert issubclass(MeanProjectionConstantSource, SourceABC)


class MeanProjectionConstantRequest(object):
    """
        Created by an MeanProjectionConstantSource to provide a way to request
        slices of the HDF5 file in a nice way.

        Note:
            This was not designed to know about dirtiness or any sort of
            changing data.

        Attributes:
            constant_request(a constant RequestABC):         The request to
                                                             take the mean
                                                             projection of.

            axis(int):                                       The axis to take
                                                             the mean
                                                             projection along.

            slicing(tuple of slices):                        Slicing to be
                                                             returned.
    """

    @prof.log_call(trace_logger)
    def __init__(self, constant_request, axis, slicing):
        """
            Constructs an MeanProjectionConstantRequest using a given file and
            path to the dataset. Optionally, throwing can be suppressed if the
            source is not found.

            Args:
                constant_request(a constant RequestABC):         The request to
                                                                 take the mean
                                                                 projection of.

                axis(int):                                       The axis to
                                                                 take the mean
                                                                 projection
                                                                 along.

                slicing(tuple of slices):                        Slicing to be
                                                                 returned.
        """

        # TODO: Look at adding assertion check on slices.

        self.constant_request = constant_request
        self.axis = axis
        self.slicing = slicing

        self._result = None

    @prof.log_call(trace_logger)
    def wait(self):
        if self._result is None:
            # Get the result of the request needed
            self._result = self.constant_request.wait()

            # Perform mean
            self._result = numpy.mean(self._result, axis=self.axis)

            # Add singleton axis where mean was performed
            self._result = xnumpy.add_singleton_axis_pos(
                self._result, self.axis
            )

            # Take the slice the viewer wanted
            self._result = self._result[self.slicing]

            logger.debug("Found the result.")

        return self._result

    @prof.log_call(trace_logger)
    def getResult(self):
        return self._result

    @prof.log_call(trace_logger)
    def cancel(self):
        pass

    @prof.log_call(trace_logger)
    def submit(self):
        pass

    # callback( result = result, **kwargs )
    @prof.log_call(trace_logger)
    def notify(self, callback, **kwargs):
        t = threading.Thread(target=self._doNotify, args=(callback, kwargs))
        t.start()

    @prof.log_call(trace_logger)
    def _doNotify(self, callback, kwargs):
        result = self.wait()
        callback(result, **kwargs)

assert issubclass(MeanProjectionConstantRequest, RequestABC)



@prof.log_call(trace_logger)
def main(*argv):
    # TODO: Try to extract code for viewing each file with each viewer. This way multiple files generates multiple viewers.

    # Only necessary if running main (normally if calling command line).
    # No point in importing otherwise.
    from nanshe.io import xjson
    import argparse

    argv = list(argv)

    # Creates command line parser
    parser = argparse.ArgumentParser(
        description="Parses input from the command line for a batch job."
    )

    # Takes a config file and then a series of one or more HDF5 files.
    parser.add_argument(
        "config_filename",
        metavar="CONFIG_FILE",
        type=str,
        help="JSON file that provides groups of items to be displayed " +
             "together with the groups to keep in sync, layer names, " +
             "and data locations."
    )
    parser.add_argument(
        "input_files",
        metavar="INPUT_FILE",
        type=str,
        nargs='+',
        help="HDF5 file(s) to use for viewing. " +
             "Must all have the same internal structure " +
             "as specified by the JSON file."
    )

    # Results of parsing arguments
    # (ignore the first one as it is the command line call).
    parsed_args = parser.parse_args(argv[1:])

    # Go ahead and stuff in parameters with the other parsed_args
    # A little risky if parsed_args may later contain a parameters variable
    # due to changing the main file or argparse changing behavior;
    # however, this keeps all arguments in the same place.
    parsed_args.parameters = xjson.read_parameters(
        parsed_args.config_filename, maintain_order=True
    )

    # Open all of the files and store their handles
    parsed_args.file_handles = []
    for i in iters.irange(len(parsed_args.input_files)):
        parsed_args.input_files[i] = parsed_args.input_files[i].rstrip("/")
        parsed_args.input_files[i] = os.path.abspath(
            parsed_args.input_files[i]
        )

        parsed_args.file_handles.append(
            h5py.File(parsed_args.input_files[i], "r")
        )

    # Make all each_layer_source_location_dict is a dict
    # whether they were or not before. The key will be the operation to
    # perform and the values will be what to perform the operation on.
    # If the key is a null string, no operation is performed.
    for i in iters.irange(len(parsed_args.parameters)):
        for (each_layer_name, each_layer_source_location_dict) in parsed_args.parameters[i].items():
            each_layer_source_operation_names = []
            each_layer_source_location_list = []

            each_layer_source_location_dict_inner = each_layer_source_location_dict
            while isinstance(each_layer_source_location_dict_inner, dict):
                assert (len(each_layer_source_location_dict_inner) == 1)
                each_layer_source_operation_names.extend(
                    list(each_layer_source_location_dict_inner.keys())
                )
                each_layer_source_location_dict_inner = list(each_layer_source_location_dict_inner.values())[0]

            if isinstance(each_layer_source_location_dict_inner, str):
                each_layer_source_location_dict_inner = [
                    each_layer_source_location_dict_inner
                ]

            assert isinstance(each_layer_source_location_dict_inner, list)

            each_layer_source_location_list = each_layer_source_location_dict_inner

            parsed_args.parameters[i][each_layer_name] = [
                each_layer_source_operation_names,
                each_layer_source_location_list
            ]

    # Find all possible matches and whether they exist or not
    parsed_args.parameters_expanded = list()
    for each_layer_names_locations_group in parsed_args.parameters:

        # As we go through all files,
        # we don't want to include the same layer more than once.
        # Also, we want to preserve the layer order in the configure file.
        parsed_args.parameters_expanded.append(collections.OrderedDict())
        for (each_layer_name, each_layer_source_location_dict) in each_layer_names_locations_group.items():

            # Fetch out the operation that we will use.
            [each_layer_source_operation_names, each_layer_source_location_list] = each_layer_source_location_dict

            # Ensure the order of files is preserved
            # (probably not an issue if they were sorted) and each is unique.
            # This way we know they get fused in the right sequential order.
            parsed_args.parameters_expanded[-1][each_layer_name] = [
                each_layer_source_operation_names, collections.OrderedDict()]
            for each_layer_source_location in each_layer_source_location_list:

                # TODO: See if we can't move this loop out. (Could change this parsed_args.parameters_expanded to an collections.OrderedDict and store none for values (ordered set).)
                for each_file in parsed_args.file_handles:
                    new_matches = hdf5.search.get_matching_grouped_paths(
                        each_file, each_layer_source_location
                    )
                    new_matches_ldict = iters.izip(
                        new_matches, itertools.repeat(None)
                    )
                    parsed_args.parameters_expanded[-1][each_layer_name][-1].update(new_matches_ldict)


    layer_names_locations_groups = parsed_args.parameters_expanded

    app = QApplication([""])#argv)
    viewer = HDF5Viewer()
    viewer.show()

    # Must reverse as Volumina puts the last items near the top.
    for each_file in reversed(parsed_args.file_handles):
        for each_layer_names_locations_group in reversed(layer_names_locations_groups):
            layer_sync_list = []

            for (each_layer_name, each_layer_source_location_dict) in reversed(
                    each_layer_names_locations_group.items()):
                [each_layer_source_operation_names, each_layer_source_location_list] = each_layer_source_location_dict

                each_source = []

                # Ignore whether the file exists as that may differ for
                # different files
                for each_layer_source_location in each_layer_source_location_list.keys():
                    each_layer_source_location = each_layer_source_location.lstrip(
                        "/")

                    each_file_source = None

                    # Try to make the source. If it fails, we take no source.
                    try:
                        if len(each_layer_source_operation_names) and\
                                ((each_layer_source_operation_names[-1].startswith("[\"") and
                                      each_layer_source_operation_names[-1].endswith("\"]")) or
                                     (each_layer_source_operation_names[-1].startswith("['") and
                                          each_layer_source_operation_names[-1].endswith("']"))):
                            each_file_source = HDF5DataSource(
                                each_file,
                                each_layer_source_location,
                                each_layer_source_operation_names.pop(-1)[2:-2]
                            )
                        else:
                            each_file_source = HDF5DataSource(
                                each_file, each_layer_source_location)
                    except HDF5DatasetNotFoundException:
                        each_file_source = None

                    each_source.append(each_file_source)

                if len(each_source) > 1:
                    try:
                        each_source = HDF5DataFusedSource(-1, *each_source)
                    except HDF5UndefinedShapeDtypeException:
                        each_source = None
                elif len(each_source) == 1:
                    each_source = each_source[0]
                else:
                    each_source = None

                for a_layer_source_operation_name in reversed(
                        each_layer_source_operation_names):
                    if (each_source is not None) and \
                            (a_layer_source_operation_name):
                        if a_layer_source_operation_name == "max":
                            each_source = MaxProjectionConstantSource(
                                each_source
                            )
                        elif a_layer_source_operation_name == "mean":
                            each_source = MeanProjectionConstantSource(
                                each_source
                            )
                        elif a_layer_source_operation_name == "enumerate":
                            each_source = EnumeratedProjectionConstantSource(
                                each_source
                            )
                        elif a_layer_source_operation_name == "contour":
                            each_source = ContourProjectionConstantSource(
                                each_source
                            )
                        elif a_layer_source_operation_name == "float":
                            each_source = FloatProjectionConstantSource(
                                each_source
                            )
                        else:
                            raise Exception(
                                "Unknown operation to perform on source \"" +
                                repr(a_layer_source_operation_name) + "\"."
                            )

                if each_source is not None:
                    each_layer = None
                    each_source_dtype = numpy.dtype(each_source.dtype()).type
                    if issubclass(each_source_dtype, numpy.integer):
                        each_layer = viewer.addColorTableHDF5Source(
                            each_source, each_source.shape(), each_layer_name
                        )
                    elif issubclass(each_source_dtype, numpy.floating):
                        each_layer = viewer.addGrayscaleHDF5Source(
                            each_source, each_source.shape(), each_layer_name
                        )
                    elif issubclass(each_source_dtype, numpy.bool_):
                        each_layer = viewer.addColorTableHDF5Source(
                            each_source, each_source.shape(), each_layer_name
                        )

                    each_layer.visible = False

                    layer_sync_list.append(each_layer)

            SyncedChannelLayers(*layer_sync_list)

    # Required to avoid segfault on Mavericks.
    # Don't change the value of app2.
    app2 = app
    exit_code = app.exec_()

    # Close and clean up files
    parsed_args.file_handles = []
    for i in iters.irange(len(parsed_args.file_handles)):
        parsed_args.file_handles[i].close()
        parsed_args.file_handles[i] = None

    parsed_args.file_handles = None
    del parsed_args.file_handles

    return(exit_code)



"""
The module ``xjson`` serializes JSON with some additional features.

===============================================================================
Overview
===============================================================================
The module ``xjson`` provides a mechanism of serializing JSON in a way that
allows for a few additional constraints.

- Commenting -- does not break the JSON specification (e.g. adding the prefix\
  ``__comment__`` to any string).
- Order dependent deserialization -- dictionary order is preserved.
- ASCII strings -- all strings are converted to ASCII

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 30, 2014 16:54:30 EDT$"


# Need in order to have logging information no matter what.
from nanshe.util import prof


try:
    unicode
except NameError:
    unicode = str


# Get the logger
trace_logger = prof.getTraceLogger(__name__)
logger = prof.logging.getLogger(__name__)


@prof.log_call(trace_logger)
def read_parameters(config_filename, maintain_order=False):
    """
        Reads the contents of a json config file and returns the parameters.

        Args:
            config_filename (str):                          name of the file
                                                            to read.

            maintain_order (bool):                          whether to preserve
                                                            the order of keys
                                                            in the json file

        Returns:
            parameters (dict or collections.OrderedDict):   parameters read
                                                            from the file.
    """


    # only relevant if reading parameter file.
    import json

    # Get the type of dictionary to use.
    json_dict = None
    if maintain_order:
        # only relevant if reading parameter file and maintaining order.
        import collections
        json_dict = collections.OrderedDict
    else:
        json_dict = dict


    @prof.log_call(trace_logger)
    def ascii_encode_str(value, json_dict=json_dict):
        """
            Encodes the str.

            Args:
                value(str):     string to encode.

            Returns:
                str:            Properly encoded str.
        """

        new_value = None
        if unicode == str and not value.startswith(u"__comment__"):
            new_value = value
        elif not value.startswith("__comment__"):
            new_value = value.encode("utf-8")

        return(new_value)


    @prof.log_call(trace_logger)
    def ascii_encode_list(data, json_dict=json_dict):
        """
            Encodes the list (and its contents).

            Args:
                data(list):      list to encode.

            Returns:
                list:            Properly encoded list.
        """
        transformed_list = []

        for each_value in data:
            new_each_value = each_value
            if isinstance(new_each_value, json_dict):
                new_each_value = ascii_encode_dict(new_each_value)
            elif isinstance(new_each_value, list):
                new_each_value = ascii_encode_list(new_each_value)
            elif isinstance(new_each_value, (bytes, unicode)):
                new_each_value = ascii_encode_str(new_each_value)

            if new_each_value is not None:
                transformed_list.append(new_each_value)

        return(transformed_list)


    @prof.log_call(trace_logger)
    def ascii_encode_dict(data, json_dict=json_dict):
        """
            Encodes the dict (and its contents).
            Also, make sure the dict is of the right type.

            Args:
                data(dict):      dict to encode.

            Returns:
                dict:            Properly encoded dict.
        """
        new_dict = json_dict(data)

        transformed_dict = []

        for each_key, each_value in new_dict.items():
            new_each_key = ascii_encode_str(each_key)

            new_each_value = each_value
            if new_each_key is not None:
                if isinstance(new_each_value, json_dict):
                    new_each_value = ascii_encode_dict(new_each_value)
                elif isinstance(new_each_value, list):
                    new_each_value = ascii_encode_list(new_each_value)
                elif isinstance(new_each_value, unicode) or \
                        isinstance(new_each_value, str):
                    new_each_value = ascii_encode_str(new_each_value)

                if new_each_value is not None:
                    transformed_dict.append((new_each_key, new_each_value))

        transformed_dict = json_dict(transformed_dict)

        return(transformed_dict)


    # gets parameters out of the file and dumps them in the dictionary. just
    # that simple.
    parameters = None
    with open(config_filename, 'r') as fp:
        logger.debug(
            "Opened configure file named \"" + config_filename + "\"."
        )

        # will just give a dictionary. just that simple
        parameters = json.load(fp, object_pairs_hook=ascii_encode_dict)

        logger.debug(
            "Loaded parameters from file, which are \"" +
            str(parameters) + "\"."
        )

    return(parameters)



"""
The module ``data`` provides a few primitives for generating synthetic data.

===============================================================================
Overview
===============================================================================
These include generating mask stacks, gaussian stacks, and random points. These
are primarily used for testing the algorithm, but could be used in benchmarking
or other applications.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Aug 01, 2014 14:55:57 EDT$"


import numpy

import scipy
import scipy.ndimage
import scipy.ndimage.filters

import nanshe.util.iters
import nanshe.util.xnumpy


def generate_hypersphere_masks(space, centers, radii, include_boundary=False):
    """
        Generate a stack of masks (first index indicates which mask); where,
        each contains a hypersphere constructed using a center and radius
        provided.

        Args:
            space(tuple of ints):                The size of the mask.

            centers(list of tuples of numbers):  List of centers with one per
                                                 hypersphere.

            radii(list of numbers):              List of radii with one per
                                                 hypersphere.

            include_boundary(bool):              Whether the mask should
                                                 contain the boundary of the
                                                 hypersphere or not.

        Returns:
            numpy.ndarray:                       A stack of masks (first index
                                                 indicates which mask) with a
                                                 filled hypersphere using a
                                                 center and radius for each
                                                 mask.

        Examples:
            >>> generate_hypersphere_masks((3, 3), (1, 1), 1.25)
            array([[[False,  True, False],
                    [ True,  True,  True],
                    [False,  True, False]]], dtype=bool)

            >>> generate_hypersphere_masks((3, 3, 3), (1, 1, 1), 1.25)
            array([[[[False, False, False],
                     [False,  True, False],
                     [False, False, False]],
            <BLANKLINE>
                    [[False,  True, False],
                     [ True,  True,  True],
                     [False,  True, False]],
            <BLANKLINE>
                    [[False, False, False],
                     [False,  True, False],
                     [False, False, False]]]], dtype=bool)

            >>> generate_hypersphere_masks((9, 9), (4, 4), 5)
            array([[[False, False,  True,  True,  True,  True,  True, False, False],
                    [False,  True,  True,  True,  True,  True,  True,  True, False],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [False,  True,  True,  True,  True,  True,  True,  True, False],
                    [False, False,  True,  True,  True,  True,  True, False, False]]], dtype=bool)

            >>> generate_hypersphere_masks(
            ...     (9, 9), (4, 4), 5, include_boundary=True
            ... )
            array([[[False,  True,  True,  True,  True,  True,  True,  True, False],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [False,  True,  True,  True,  True,  True,  True,  True, False]]], dtype=bool)
    """

    return(generate_hyperdisc_masks(space, centers, radii, include_boundary))


def generate_hyperdisc_masks(space,
                             centers,
                             radii_max,
                             include_boundary_max=False,
                             radii_min=None,
                             include_boundary_min=False):
    """
        Generate a stack of masks (first index indicates which mask); where,
        each contains a hyperdisc constructed using a center and radius
        provided.

        Args:
            space(tuple of ints):                The size of the mask.

            centers(list of tuples of numbers):  List of centers with one per
                                                 hyperdisc.

            radii_max(list of numbers):          List of max radii with one per
                                                 hyperdisc.

            include_boundary_max(bool):          Whether the mask should
                                                 contain the max boundary of
                                                 the hyperdisc or not.

            radii_min(list of numbers):          List of min radii with one per
                                                 hyperdisc.

            include_boundary_min(bool):          Whether the mask should
                                                 contain the min boundary of
                                                 the hyperdisc or not.

        Returns:
            numpy.ndarray:                       A stack of masks (first index
                                                 indicates which mask) with a
                                                 filled hyperdisc using a
                                                 center and radius for each
                                                 mask.

        Examples:
            >>> generate_hyperdisc_masks((3, 3), (1, 1), 1.25)
            array([[[False,  True, False],
                    [ True,  True,  True],
                    [False,  True, False]]], dtype=bool)

            >>> generate_hyperdisc_masks((3, 3), (1, 1), 1.25, radii_min=0.0, include_boundary_min=False)
            array([[[False,  True, False],
                    [ True, False,  True],
                    [False,  True, False]]], dtype=bool)

            >>> generate_hyperdisc_masks((3, 3, 3), (1, 1, 1), 1.25)
            array([[[[False, False, False],
                     [False,  True, False],
                     [False, False, False]],
            <BLANKLINE>
                    [[False,  True, False],
                     [ True,  True,  True],
                     [False,  True, False]],
            <BLANKLINE>
                    [[False, False, False],
                     [False,  True, False],
                     [False, False, False]]]], dtype=bool)

            >>> generate_hyperdisc_masks((9, 9), (4, 4), 5)
            array([[[False, False,  True,  True,  True,  True,  True, False, False],
                    [False,  True,  True,  True,  True,  True,  True,  True, False],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [False,  True,  True,  True,  True,  True,  True,  True, False],
                    [False, False,  True,  True,  True,  True,  True, False, False]]], dtype=bool)

            >>> generate_hyperdisc_masks(
            ...     (9, 9), (4, 4), 5, include_boundary_max=True
            ... )
            array([[[False,  True,  True,  True,  True,  True,  True,  True, False],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True,  True,  True,  True],
                    [False,  True,  True,  True,  True,  True,  True,  True, False]]], dtype=bool)

            >>> generate_hyperdisc_masks(
            ...     (9, 9), (4, 4), 5, True, 4, True
            ... )
            array([[[False,  True,  True,  True,  True,  True,  True,  True, False],
                    [ True,  True, False, False, False, False, False,  True,  True],
                    [ True, False, False, False, False, False, False, False,  True],
                    [ True, False, False, False, False, False, False, False,  True],
                    [ True, False, False, False, False, False, False, False,  True],
                    [ True, False, False, False, False, False, False, False,  True],
                    [ True, False, False, False, False, False, False, False,  True],
                    [ True,  True, False, False, False, False, False,  True,  True],
                    [False,  True,  True,  True,  True,  True,  True,  True, False]]], dtype=bool)
    """

    # Convert to arrays
    space = numpy.array(space)
    centers = numpy.array(centers)
    radii_max = numpy.array(radii_max)
    if radii_min is not None:
        radii_min = numpy.array(radii_min)

    # Add a singleton dimension if there is only one of each.
    if centers.ndim == 1:
        centers = centers[None]

    if radii_max.ndim == 0:
        radii_max = radii_max[None]

    if (radii_min is not None) and (radii_min.ndim == 0):
            radii_min = radii_min[None]

    # Validate the dimensions
    assert (space.ndim == 1)
    assert (centers.ndim == 2)
    assert (radii_max.ndim == 1)
    if radii_min is not None:
        assert (radii_min.ndim == 1)

    # Validate the shapes
    assert (space.shape == centers.shape[1:])
    assert (radii_max.shape == centers.shape[:1])
    if radii_min is not None:
        assert (radii_min.shape == centers.shape[:1])

    radii_max_2 = radii_max**2
    radii_max_2 = radii_max_2[(Ellipsis,) + (None,)*len(space)]
    if radii_min is not None:
        radii_min_2 = radii_min**2
        radii_min_2 = radii_min_2[(Ellipsis,) + (None,)*len(space)]

    space_indices = numpy.indices(space)[None]

    # Create a hyperdisc mask using a center and a radius.
    hyperdisc_masks = None

    point_offsets_dist_2 = (
        (space_indices - centers[(Ellipsis,) + (None,)*len(space)])**2
    ).sum(axis=1)

    if include_boundary_max:
        hyperdisc_masks = (point_offsets_dist_2 <= radii_max_2)
    else:
        hyperdisc_masks = (point_offsets_dist_2 < radii_max_2)
    if radii_min is not None:
        if include_boundary_min:
            hyperdisc_masks &= (radii_min_2 <= point_offsets_dist_2)
        else:
            hyperdisc_masks &= (radii_min_2 < point_offsets_dist_2)

    return(hyperdisc_masks)


def generate_gaussian_images(space, means, std_devs, magnitudes):
    """
        Generate a stack of gaussians (first index indicates which image);
        where, each contains a gaussian using a center and radius provided.

        Note:
            This uses a normalized gaussian filter to create the gaussians. So,
            the sum over any image in the stack will be the magnitude given
            (with some small error).

        Args:
            space(tuple of ints):                The size of the mask.

            means(list of tuples of numbers):    List of means with one per
                                                 gaussian.

            std_devs(list of numbers):           List of standard deviations
                                                 with one per gaussian.

            magnitudes(list of numbers):         List of magnitudes for each
                                                 gaussian.

        Returns:
            numpy.ndarray:                       A stack of masks (first index
                                                 indicates which mask) with a
                                                 gaussian using a center and
                                                 radius for each mask.

        Examples:
            >>> generate_gaussian_images((5, 5), (2, 2), 0.5, 1) #doctest: +NORMALIZE_WHITESPACE
            array([[[ 6.96247819e-08, 2.80886418e-05, 2.07548550e-04, 2.80886418e-05, 6.96247819e-08],
                    [ 2.80886418e-05, 1.13317669e-02, 8.37310610e-02, 1.13317669e-02, 2.80886418e-05],
                    [ 2.07548550e-04, 8.37310610e-02, 6.18693507e-01, 8.37310610e-02, 2.07548550e-04],
                    [ 2.80886418e-05, 1.13317669e-02, 8.37310610e-02, 1.13317669e-02, 2.80886418e-05],
                    [ 6.96247819e-08, 2.80886418e-05, 2.07548550e-04, 2.80886418e-05, 6.96247819e-08]]])

            >>> round(
            ...     generate_gaussian_images((5, 5), (2, 2), 0.5, 1).sum(), 3
            ... )
            1.0

            >>> generate_gaussian_images((5, 5), (2, 2), 0.25, 1) #doctest: +NORMALIZE_WHITESPACE
            array([[[ 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00, 1.12384321e-07, 3.35012940e-04, 1.12384321e-07, 0.00000000e+00],
                    [ 0.00000000e+00, 3.35012940e-04, 9.98659499e-01, 3.35012940e-04, 0.00000000e+00],
                    [ 0.00000000e+00, 1.12384321e-07, 3.35012940e-04, 1.12384321e-07, 0.00000000e+00],
                    [ 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]]])

            >>> round(
            ...     generate_gaussian_images((5, 5), (2, 2), 0.25, 1).sum(), 3
            ... )
            1.0

            >>> generate_gaussian_images(
            ...     (5, 5), (2, 2), 0.25, 5
            ... ) #doctest: +NORMALIZE_WHITESPACE
            array([[[ 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00, 5.61921606e-07, 1.67506470e-03, 5.61921606e-07, 0.00000000e+00],
                    [ 0.00000000e+00, 1.67506470e-03, 4.99329749e+00, 1.67506470e-03, 0.00000000e+00],
                    [ 0.00000000e+00, 5.61921606e-07, 1.67506470e-03, 5.61921606e-07, 0.00000000e+00],
                    [ 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00]]])

            >>> round(
            ...     generate_gaussian_images((5, 5), (2, 2), 0.25, 5).sum(), 3
            ... )
            5.0

            >>> generate_gaussian_images((3, 3, 3), (1, 1, 1), 0.25, 100)
            array([[[[  3.76754623e-09,   1.12308970e-05,   3.76754623e-09],
                     [  1.12308970e-05,   3.34788322e-02,   1.12308970e-05],
                     [  3.76754623e-09,   1.12308970e-05,   3.76754623e-09]],
            <BLANKLINE>
                    [[  1.12308970e-05,   3.34788322e-02,   1.12308970e-05],
                     [  3.34788322e-02,   9.97989922e+01,   3.34788322e-02],
                     [  1.12308970e-05,   3.34788322e-02,   1.12308970e-05]],
            <BLANKLINE>
                    [[  3.76754623e-09,   1.12308970e-05,   3.76754623e-09],
                     [  1.12308970e-05,   3.34788322e-02,   1.12308970e-05],
                     [  3.76754623e-09,   1.12308970e-05,   3.76754623e-09]]]])

            >>> round(
            ...     generate_gaussian_images(
            ...         (3, 3, 3), (1, 1, 1), 0.25, 100
            ...     ).sum(),
            ...     3
            ... )
            100.0
    """

    # Convert to arrays
    space = numpy.array(space)
    means = numpy.array(means)
    std_devs = numpy.array(std_devs)
    magnitudes = numpy.array(magnitudes)

    # Add a singleton dimension if there is only one of each.
    if means.ndim == 1:
        means = means[None]

    if std_devs.ndim == 0:
        std_devs = std_devs[None]

    if magnitudes.ndim == 0:
        magnitudes = magnitudes[None]

    # Validate the dimensions
    assert (space.ndim == 1)
    assert (means.ndim == 2)
    assert (std_devs.ndim == 1)
    assert (magnitudes.ndim == 1)

    # Validate the shapes
    assert (space.shape == means.shape[1:])
    assert (magnitudes.shape == means.shape[:1])

    # Create a gaussian from a mean and a standard deviation.
    images = numpy.zeros(
        magnitudes.shape + tuple(space.tolist()), dtype=float
    )
    for i, (each_mean, each_std_dev, each_magnitude) in enumerate(
            nanshe.util.iters.izip(means, std_devs, magnitudes)
    ):
        images[i][tuple(each_mean)] = each_magnitude
        images[i] = scipy.ndimage.filters.gaussian_filter(
            images[i],
            each_std_dev,
            mode="nearest"
        )

    return(images)


def generate_random_bound_points(space, radii):
    """
        Generate a collection of random points that are the center of a
        hypersphere contained within the space.

        Note:
            Uses the typical convention [0, space_i) for all space_i from
            space. Applies the additional constraint
            [radius_i, space_i - radius_i).

        Args:
            space(tuple of ints):                The size of the space where
                                                 the points will lie.

            radii(list of numbers):              Additional radius for each
                                                 point to remain away from the
                                                 boundary.

        Returns:
            numpy.ndarray:                       An array with the random
                                                 points each using a radius
                                                 from radii to constrain
                                                 itself. First index is which
                                                 point. Second is which
                                                 coordinate.

        Examples:
            >>> numpy.random.seed(0)
            >>> generate_random_bound_points((100, 100), 5)
            array([[49, 52]])

            >>> numpy.random.seed(0)
            >>> generate_random_bound_points((100, 100), (5, 5))
            array([[49, 52],
                   [69, 72]])

            >>> numpy.random.seed(0)
            >>> generate_random_bound_points((100, 100, 100), (5,))
            array([[49, 52, 69]])

            >>> numpy.random.seed(0)
            >>> generate_random_bound_points((100, 100, 100), (5, 5))
            array([[49, 52, 69],
                   [72, 72, 14]])
    """

    # Convert to arrays
    space = numpy.array(space)
    radii = numpy.array(radii)

    # Add a singleton dimension if there is only one of each.
    if radii.ndim == 0:
        radii = radii[None]

    # Validate the dimensions
    assert (space.ndim == 1)
    assert (radii.ndim == 1)

    # Determine the space each centroid can be within
    bound_space = numpy.zeros(radii.shape + space.shape + (2,), dtype=int)
    bound_space[..., 0] = nanshe.util.xnumpy.expand_view(radii, space.shape)
    bound_space[..., 1] = nanshe.util.xnumpy.expand_view(space, reps_before=radii.shape) - \
                          nanshe.util.xnumpy.expand_view(radii, space.shape)

    # Generate a random point for each radius.
    points = numpy.zeros(radii.shape + space.shape, dtype=int)
    for i in nanshe.util.iters.irange(len(radii)):
        for j in nanshe.util.iters.irange(len(space)):
            points[i][j] = numpy.random.randint(
                bound_space[i][j][0], bound_space[i][j][1]
            )

    return(points)



"""
The ``wavelet`` module provides support for performing a wavelet transform.

===============================================================================
Overview
===============================================================================
Included are tools to aid in the computation of the wavelet transform. In
particular, construction of the kernel at different scales and it application
to the data. The kernel applied is based on the technique presented by
Reichinnek, et al. ( doi:`10.1016/j.neuroimage.2011.12.018`_ ).


.. _`10.1016/j.neuroimage.2011.12.018`: \
    http://dx.doi.org/10.1016/j.neuroimage.2011.12.018

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$May 01, 2014 14:24:55 EDT$"


import warnings

import numpy

import vigra

from nanshe.io import hdf5
from nanshe.util.iters import irange
from nanshe.util.xnumpy import binomial_coefficients


# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def binomial_1D_array_kernel(i, n=4):
    """
        Generates a 1D numpy array used to make the kernel for the wavelet
        transform.

        Args:
            i(int):               which scaling to use.
            n(int):               which row of Pascal's triangle to return.

        Returns:
            r(numpy.ndarray):     a 1D numpy array to use as the wavelet
                                  transform kernel.


        Examples:
            >>> binomial_1D_array_kernel(0, -2)
            array([], dtype=float64)

            >>> binomial_1D_array_kernel(0, 0)
            array([ 1.])

            >>> binomial_1D_array_kernel(0)
            array([ 0.0625,  0.25  ,  0.375 ,  0.25  ,  0.0625])

            >>> binomial_1D_array_kernel(0, 4)
            array([ 0.0625,  0.25  ,  0.375 ,  0.25  ,  0.0625])

            >>> binomial_1D_array_kernel(1, 4)
            array([ 0.0625,  0.25  ,  0.375 ,  0.25  ,  0.0625])

            >>> binomial_1D_array_kernel(2, 4)
            array([ 0.0625,  0.    ,  0.25  ,  0.    ,  0.375 ,  0.    ,  0.25  ,
                    0.    ,  0.0625])

            >>> binomial_1D_array_kernel(3, 4)
            array([ 0.0625,  0.    ,  0.    ,  0.    ,  0.25  ,  0.    ,  0.    ,
                    0.    ,  0.375 ,  0.    ,  0.    ,  0.    ,  0.25  ,  0.    ,
                    0.    ,  0.    ,  0.0625])

            >>> binomial_1D_array_kernel(4, 4)
            array([ 0.0625,  0.    ,  0.    ,  0.    ,  0.    ,  0.    ,  0.    ,
                    0.    ,  0.25  ,  0.    ,  0.    ,  0.    ,  0.    ,  0.    ,
                    0.    ,  0.    ,  0.375 ,  0.    ,  0.    ,  0.    ,  0.    ,
                    0.    ,  0.    ,  0.    ,  0.25  ,  0.    ,  0.    ,  0.    ,
                    0.    ,  0.    ,  0.    ,  0.    ,  0.0625])

            >>> binomial_1D_array_kernel(2, 1)
            array([ 0.5,  0. ,  0.5])
    """

    # Below 1 is irrelevant.
    if i < 1:
        i = 1

    # Get the binomial coefficients.
    cs = list(binomial_coefficients(n))

    # Reuse the correction to `n` found by `binomial_coefficients`.
    n = len(cs) - 1

    # Get the right number of zeros to fill in
    zs = list(numpy.zeros(2 ** (i - 1) - 1, dtype=int))

    # Create the contents of the 1D kernel before normalization
    r = []
    if len(cs) > 1:
        for _ in cs[:-1]:
            r.append(_)
            r.extend(zs)

        r.append(cs[-1])
    else:
        r.extend(cs)

    r = numpy.array(r)
    r = r.astype(float)

    # Normalization on the L_1 norm.
    r /= 2 ** n

    return(r)


@prof.log_call(trace_logger)
def binomial_1D_vigra_kernel(i, n=4, border_treatment=vigra.filters.BorderTreatmentMode.BORDER_TREATMENT_REFLECT):
    """
        Generates a vigra.filters.Kernel1D using binomial_1D_array_kernel(i).

        Args:
            i(int):                                                 which
                                                                    scaling to
                                                                    use.

            n(int):                                                 which row
                                                                    of Pascal's
                                                                    triangle to
                                                                    return.

            border_treatment(vigra.filters.BorderTreatmentMode):    determines
                                                                    how to deal
                                                                    with the
                                                                    borders.

        Returns:
            k(vigra.filters.Kernel1D):                              a 1D vigra
                                                                    kernel to
                                                                    aid in
                                                                    computing
                                                                    the wavelet
                                                                    transform.


        Examples:
            >>> binomial_1D_vigra_kernel(1) # doctest: +ELLIPSIS
            <vigra.filters.Kernel1D object at 0x...>
    """

    # Generate the vector for the kernel
    h_kern = binomial_1D_array_kernel(i, n)

    # Determine the kernel center
    h_kern_half_width = (h_kern.size - 1) // 2

    # Default kernel
    k = vigra.filters.Kernel1D()
    # Center the kernel
    k.initExplicitly(-h_kern_half_width, h_kern_half_width, h_kern)
    # Set the border treatment mode.
    k.setBorderTreatment(border_treatment)

    return(k)


@prof.log_call(trace_logger)
@hdf5.record.static_array_debug_recorder
def transform(im0,
              scale=5,
              include_intermediates=False,
              include_lower_scales=False,
              out=None):
    """
        Performs integral steps of the wavelet transform on im0 up to the given
        scale. If scale is an iterable, then

        Args:
            im0(numpy.ndarray):                  the original image.
            scale(int or tuple of ints):         the scale of wavelet transform
                                                 to apply.

            include_intermediates(bool):         whether to return
                                                 intermediates or not
                                                 (default False).

            include_lower_scales(bool):          whether to include lower
                                                 scales or not (default False)
                                                 (ignored if
                                                 include_intermediates is True)

            out(numpy.ndarray):                  holds final result (cannot use
                                                 unless include_intermediates
                                                 is False or an AssertionError
                                                 will be raised.)

        Returns:
            W, out(tuple of numpy.ndarrays):     returns the final result of
                                                 the wavelet transform and
                                                 possibly other scales. Also,
                                                 may return the intermediates.


        Examples:
            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = True,
            ...     include_lower_scales = True) # doctest: +NORMALIZE_WHITESPACE
            (array([[[ 0.59375, -0.375  , -0.34375],
                     [-0.375  ,  0.625  , -0.375  ],
                     [-0.34375, -0.375  ,  0.59375]]], dtype=float32),
             array([[[ 1.     ,  0.     ,  0.     ],
                     [ 0.     ,  1.     ,  0.     ],
                     [ 0.     ,  0.     ,  1.     ]],
                    [[ 0.40625,  0.375  ,  0.34375],
                     [ 0.375  ,  0.375  ,  0.375  ],
                     [ 0.34375,  0.375  ,  0.40625]]], dtype=float32))

            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = True)
            array([[[ 0.59375, -0.375  , -0.34375],
                    [-0.375  ,  0.625  , -0.375  ],
                    [-0.34375, -0.375  ,  0.59375]]], dtype=float32)

            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = False)
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)

            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = (0, 1),
            ...     include_intermediates = False,
            ...     include_lower_scales = False)
            array([[ 0.625, -0.25 , -0.125],
                   [-0.5  ,  0.5  , -0.5  ],
                   [-0.125, -0.25 ,  0.625]], dtype=float32)

            >>> out = numpy.zeros((3, 3), dtype = numpy.float32)
            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = False,
            ...     out = out)
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)
            >>> out
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)

            >>> out = numpy.eye(3, dtype = numpy.float32)
            >>> transform(out,
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = False,
            ...     out = out)
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)
            >>> out
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)

            >>> out = numpy.empty((1, 3, 3), dtype = numpy.float32)
            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = True,
            ...     out = out) # doctest: +NORMALIZE_WHITESPACE
            array([[[ 0.59375, -0.375  , -0.34375],
                    [-0.375  ,  0.625  , -0.375  ],
                    [-0.34375, -0.375  ,  0.59375]]], dtype=float32)
            >>> out
            array([[[ 0.59375, -0.375  , -0.34375],
                    [-0.375  ,  0.625  , -0.375  ],
                    [-0.34375, -0.375  ,  0.59375]]], dtype=float32)

            >>> out = numpy.empty((1, 3, 3), dtype = numpy.float64)
            >>> transform(numpy.eye(3, dtype = numpy.float32),
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = True,
            ...     out = out) # doctest: +NORMALIZE_WHITESPACE
            array([[[ 0.59375, -0.375  , -0.34375],
                    [-0.375  ,  0.625  , -0.375  ],
                    [-0.34375, -0.375  ,  0.59375]]])
            >>> out
            array([[[ 0.59375, -0.375  , -0.34375],
                    [-0.375  ,  0.625  , -0.375  ],
                    [-0.34375, -0.375  ,  0.59375]]])

            >>> out = numpy.eye(3, dtype = numpy.uint8)
            >>> transform(out,
            ...     scale = 1,
            ...     include_intermediates = False,
            ...     include_lower_scales = False,
            ...     out = out)
            array([[ 0.59375, -0.375  , -0.34375],
                   [-0.375  ,  0.625  , -0.375  ],
                   [-0.34375, -0.375  ,  0.59375]], dtype=float32)
            >>> out
            array([[1, 0, 0],
                   [0, 1, 0],
                   [0, 0, 1]], dtype=uint8)
    """

    if not issubclass(im0.dtype.type, numpy.float32):
        warnings.warn(
            "Provided im0 with type \"" + repr(im0.dtype.type) + "\". " +
            "Will be cast to type \"" + repr(numpy.float32) + "\"",
            RuntimeWarning
        )

        im0 = im0.astype(numpy.float32)

    # Make sure that we have scale as a list.
    # If it is not a list, then make a singleton list.
    try:
        scale = numpy.array(list(scale))

        assert (scale.ndim == 1), \
            "Scale should only have 1 dimension. " + \
            "Instead, got scale.ndim = \"" + str(scale.ndim) + "\"."

        assert (len(scale) == im0.ndim), \
            "Scale should have a value of each dimension of im0. " + \
            "Instead, got len(scale) = \"" + str(len(scale)) + "\" and " + \
            "im0.ndim = \"" + str(im0.ndim) + "\"."

    except TypeError:
        scale = numpy.repeat([scale], im0.ndim)


    imPrev = None
    imCur = None
    if include_intermediates:
        assert (out is None)

        W = numpy.zeros((scale.max(),) + im0.shape, dtype=numpy.float32)
        imOut = numpy.zeros(
            (scale.max() + 1,) + im0.shape, dtype=numpy.float32
        )
        imOut[0] = im0

        imCur = imOut[0]
        imPrev = imCur
    else:
        if include_lower_scales:
            if out is None:
                W = numpy.zeros(
                    (scale.max(),) + im0.shape, dtype=numpy.float32
                )
                out = W
            else:
                assert (out.shape == ((scale.max(),) + im0.shape))

                if not issubclass(out.dtype.type, numpy.float32):
                    warnings.warn(
                        "Provided out with type \"" + repr(out.dtype.type) +
                        "\". " +
                        "Will be cast to type \"" + repr(numpy.float32) + "\"",
                        RuntimeWarning
                    )

                W = out

            imPrev = numpy.empty_like(im0)
        else:
            if out is not None:
                assert (out.shape == im0.shape)

                if not issubclass(out.dtype.type, numpy.float32):
                    warnings.warn(
                        "Provided out with type \"" + repr(out.dtype.type) +
                        "\". " +
                        "Will be cast to type \"" + repr(numpy.float32) + "\"",
                        RuntimeWarning
                    )

                    out = im0.astype(numpy.float32)

                imPrev = out
            else:
                imPrev = numpy.empty_like(im0)
                out = imPrev

        imCur = im0.astype(numpy.float32)


    for i in irange(1, scale.max() + 1):
        if include_intermediates:
            imPrev = imCur
            imOut[i] = imOut[i - 1]
            imCur = imOut[i]
        else:
            imPrev[:] = imCur

        h_ker = binomial_1D_vigra_kernel(i)

        for d in irange(len(scale)):
            if i <= scale[d]:
                vigra.filters.convolveOneDimension(imCur, d, h_ker, out=imCur)

        if include_intermediates or include_lower_scales:
            W[i - 1] = imPrev - imCur

    if include_intermediates:
        return((W, imOut))
    elif include_lower_scales:
        return(W)
    else:
        # Same as returning imPrev - imCur.
        # Except, it avoids generating another array to hold the result.
        out -= imCur
        return(out)



"""
The module ``wrappers`` provides support decorating functions and classes.

===============================================================================
Overview
===============================================================================
The module ``wrappers`` extends wrapping abilities found in |functools|_. In
particular, it is ensured all wrapped functions contain an attribute
``__wrapped__``, which points back to the original function before the wrapper
was applied. Also, the ability to wrap classes with a decorator to apply a
``metaclass`` or series of ``metaclass``es is provided. Making it much easier
to transform classes without mucking in their internals. For classes inheriting
from ``Qt`` objects a special decorator is provided to make sure they
participate in the correct inheritance scheme.

.. |functools| replace:: ``functools``
.. _functools: http://docs.python.org/2/library/functools.html

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jul 23, 2014 16:24:36 EDT$"


import collections
import inspect
import functools
import types

import PyQt4.QtCore


def update_wrapper(wrapper,
                   wrapped,
                   assigned=functools.WRAPPER_ASSIGNMENTS,
                   updated=functools.WRAPPER_UPDATES):
    """
        Extends functools.update_wrapper to ensure that it stores the wrapped
        function in the attribute __wrapped__.

        Args:
            wrapper(callable):      the replacement callable.

            wrapped(callable):      the callable that is being wrapped.

            assigned(tuple):        is a tuple naming the attributes assigned
                                    directly from the wrapped function to the
                                    wrapper function (defaults to
                                    functools.WRAPPER_ASSIGNMENTS)

            updated(tuple):         is a tuple naming the attributes of the
                                    wrapper that are updated with the
                                    corresponding attribute from the wrapped
                                    function (defaults to
                                    functools.WRAPPER_UPDATES)

        Returns:
            (callable):             the wrapped callable.
    """

    wrapper = functools.update_wrapper(
        wrapper, wrapped, assigned=assigned, updated=updated
    )

    # Store the underlying callable. Automatic in Python 3.
    setattr(wrapper, "__wrapped__", getattr(wrapper, "__wrapped__", wrapped))

    return(wrapper)


def wraps(wrapped,
          assigned=functools.WRAPPER_ASSIGNMENTS,
          updated=functools.WRAPPER_UPDATES):
    """
        Builds on functools.wraps to ensure that it stores the wrapped function
        in the attribute __wrapped__.

        Args:
            wrapped(callable):      the callable that is being wrapped.

            assigned(tuple):        is a tuple naming the attributes assigned
                                    directly from the wrapped function to the
                                    wrapper function (defaults to
                                    functools.WRAPPER_ASSIGNMENTS)

            updated(tuple):         is a tuple naming the attributes of the
                                    wrapper that are updated with the
                                    corresponding attribute from the wrapped
                                    function (defaults to
                                    functools.WRAPPER_UPDATES)

        Returns:
            (callable):             a decorator for callable, which will
                                    contain wrapped.
    """

    return(functools.partial(
        update_wrapper, wrapped=wrapped, assigned=assigned, updated=updated
    ))


def identity_wrapper(a_callable):
    """
        Trivially wraps a given callable without doing anything else to it.

        Args:
            a_callable(callable):   the callable that is being wrapped.

        Returns:
            (callable):             a wrapped callable.
    """

    @wraps(a_callable)
    def wrapped_callable(*args, **kwargs):
        """
            Trivially wraps a given callable without doing anything else to it.

            Args:
                *args:      Variable length argument list.
                **kwargs:   Arbitrary keyword arguments.

            Returns:
                Same as what `a_callable` returns.
        """

        return(a_callable(*args, **kwargs))

    return(wrapped_callable)


def static_variables(**kwargs):
    """
        Returns a decorator that decorates a callable such that it has the
        given static variables set.

        Args:
            *kwargs(tuple):     keyword args will be set to the value provided.

        Returns:
            (decorator):        a decorator for the callable.
    """

    def static_variables_tie(a_callable):
        """
            Decorates a function such that it has the given static variables
            set.

            Args:
                a_callable(callable):     the callable to decorate.

            Returns:
                (callable):               the callable returned.

        """

        callable_wrapped = identity_wrapper(a_callable)

        for each_kwd, each_val in kwargs.items():
            setattr(callable_wrapped, each_kwd, each_val)

        return(callable_wrapped)

    return(static_variables_tie)


def metaclass(meta):
    """
        Returns a decorator that decorates a class such that the given
        metaclass is applied.

        Note:
            Decorator will add the __metaclass__ attribute so the last
            metaclass applied is known. Also, decorator will add the
            __wrapped__ attribute so that the unwrapped class can be retrieved.

        Args:
            meta(metaclass):     metaclass to apply to a given class.

        Returns:
            (decorator):         a decorator for the class.
    """

    def metaclass_wrapper(cls):
        """
            Returns a decorated class such that the given metaclass is applied.

            Note:
                Adds the __metaclass__ attribute so the last metaclass used is
                known. Also, adds the __wrapped__ attribute so that the
                unwrapped class can be retrieved.

            Args:
                cls(class):          class to decorate.

            Returns:
                (class):             the decorated class.

        """

        __name = str(cls.__name__)
        __bases = tuple(cls.__bases__)
        __dict = dict(cls.__dict__)

        __dict.pop("__dict__", None)
        __dict.pop("__weakref__", None)

        for each_slot in __dict.get("__slots__", tuple()):
            __dict.pop(each_slot, None)

        __dict["__metaclass__"] = meta
        __dict["__wrapped__"] = cls

        return(meta(__name, __bases, __dict))

    return(metaclass_wrapper)


def metaclasses(*metas):
    """
        Returns a decorator that decorates a class such that the given
        metaclasses are applied.

        Note:
            Shorthand for repeated application of metaclass.

        Args:
            *metas(metaclasses):     metaclasses to apply to a given class.

        Returns:
            (decorator):             a decorator for the class.
    """

    def metaclasses_wrapper(cls):
        """
            Returns a decorated class such that the given metaclasses are
            applied.

            Args:
                cls(class):          class to decorate.

            Returns:
                (class):             the decorated class.

        """

        new_cls = cls

        for each_meta in metas:
            new_cls = metaclass(each_meta)(new_cls)

        return(new_cls)

    return(metaclasses_wrapper)


def class_static_variables(**kwargs):
    """
        Returns a decorator that decorates a class such that it has the given
        static variables set.

        Args:
            **kwargs(tuple):     keyword args will be set to the value
                                 provided.

        Returns:
            (decorator):         a decorator for the class.
    """

    class MetaStaticVariables(type):
        """
            Metaclass, which adds static variable with the given value to a
            class.
        """

        def __new__(meta, name, bases, dct):
            dct.update(kwargs)

            return(super(MetaStaticVariables, meta).__new__(
                meta, name, bases, dct
            ))

    return(metaclass(MetaStaticVariables))


def class_decorate_all_methods(*decorators):
    """
        Returns a decorator that decorates a class such that all its methods
        are decorated by the decorators provided.

        Args:
            *decorators(tuple):     decorators to decorate all methods with.

        Returns:
            (decorator):            a decorator for the class.
    """

    class MetaAllMethodsDecorator(type):
        """
            Metaclass, which decorates all methods with the list of decorators
            in order.
        """

        def __new__(meta, name, bases, dct):
            for _k, _v in dct.items():
                # Are all of FunctionType at this point.
                # Will be of MethodType at a later step.
                if isinstance(_v, types.FunctionType):
                    for each_decorator in decorators:
                        _v = each_decorator(_v)

                dct[_k] = _v

            return(super(MetaAllMethodsDecorator, meta).__new__(
                meta, name, bases, dct
            ))

    return(metaclass(MetaAllMethodsDecorator))


def qt_class_decorate_all_methods(*decorators):
    """
        Returns a decorator that decorates a class such that all its methods
        are decorated by the decorators provided.

        Args:
            *decorators(tuple):     decorators to decorate all methods with.

        Returns:
            (decorator):            a decorator for the class.
    """

    class MetaAllMethodsDecorator(PyQt4.QtCore.pyqtWrapperType):
        """
            Metaclass, which decorates all methods with the list of decorators
            in order.

            Inherits from PyQt4.QtCore.pyqtWrapperType based on this
            ( http://www.gulon.co.uk/2012/12/28/pyqt4-qobjects-and-metaclasses/ ).
        """

        def __new__(meta, name, bases, dct):
            for _k, _v in dct.items():
                # Are all of FunctionType at this point.
                # Will be of MethodType at a later step.
                if isinstance(_v, types.FunctionType):
                    for each_decorator in decorators:
                        _v = each_decorator(_v)

                dct[_k] = _v

            return(super(MetaAllMethodsDecorator, meta).__new__(
                meta, name, bases, dct
            ))

    return(metaclass(MetaAllMethodsDecorator))


def class_decorate_methods(**method_decorators):
    """
        Returns a decorator that decorates a class such that specified methods
        are decorated by the decorators provided.

        Args:
            **method_decorators(tuple):     method names with a single
                                            decorator or a list of decorators.

        Returns:
            (decorator):                    a decorator for the class.
    """

    class MetaMethodsDecorator(type):
        """
            Metaclass, which decorates some methods based on the keys given.
            Uses the decorator(s) provided for each method to decorator in
            order.
        """

        def __new__(meta, name, bases, dct):
            for _k, _v in dct.items():
                if isinstance(_v, types.FunctionType):
                    _dl = method_decorators.get(_k)
                    if (_dl is not None):
                        try:
                            iter(_dl)
                        except TypeError:
                            _dl = [_dl]

                        for _d in _dl:
                            _v = _d(_v)

                dct[_k] = _v

            return(super(MetaMethodsDecorator, meta).__new__(
                meta, name, bases, dct
            ))

    return(metaclass(MetaMethodsDecorator))


def unwrap(a_callable):
    """
        Returns the underlying function that was wrapped.

        Args:
            a_callable(callable):     some wrapped (or not) callable.

        Returns:
            (callable):               the callable that is no longer wrapped.
    """

    unwrapped_callable = a_callable

    while hasattr(unwrapped_callable, "__wrapped__"):
        unwrapped_callable = unwrapped_callable.__wrapped__

    return(unwrapped_callable)


def tied_call_args(a_callable, *args, **kwargs):
    """
        Ties all the args to their respective variable names.

        Args:
            a_callable(callable):     some callable.
            *args(callable):          positional arguments for the callable.
            **kwargs(callable):       keyword arguments for the callable.

        Returns:
            args (tuple):             ordered dictionary of arguments name and
                                      their values, all variadic position
                                      arguments, all variadic keyword
                                      arguments.
    """

    sig = inspect.getargspec(a_callable)

    unsorted_callargs = inspect.getcallargs(a_callable, *args, **kwargs)

    new_args = tuple()
    if (sig.varargs is not None):
        new_args = unsorted_callargs[sig.varargs]

    new_kwargs = dict()
    if (sig.keywords is not None):
        new_kwargs = unsorted_callargs[sig.keywords]

    callargs = collections.OrderedDict()
    for each_arg in sig.args:
        callargs[each_arg] = unsorted_callargs[each_arg]

    return(callargs, new_args, new_kwargs)


def repack_call_args(a_callable, *args, **kwargs):
    """
        Reorganizes args and kwargs to match the given callables signature.

        Args:
            a_callable(callable):     some callable.
            *args(callable):          positional arguments for the callable.
            **kwargs(callable):       keyword arguments for the callable.

        Returns:
            args (tuple):             all arguments as passed as position
                                      arguments, all default arguments and
                                      all arguments passed as keyword
                                      arguments.
    """

    callargs, new_args, new_kwargs = tied_call_args(
        a_callable, *args, **kwargs
    )

    new_args = tuple(list(callargs.values())[:len(args)]) + new_args
    new_kwargs.update(dict(list(callargs.items())[len(args):]))

    return(new_args, new_kwargs)


def with_setup_state(setup=None, teardown=None):
    """
        Adds setup and teardown callable to a function s.t. they can mutate it.

        Based on ``with_setup`` from ``nose``. This goes a bit further than
        ``nose`` does and provides a mechanism for the setup and teardown
        functions to change the callable in question. In other words, variables
        generated in setup can be stored in the functions globals and then
        cleaned up and removed in teardown. The final result of using this
        function should be a function equivalent to one generated by
        ``with_setup``.

        Args:
            setup(callable):        A callable that takes the decorated
                                    function as an argument. This sets up the
                                    function before execution.

            teardown(callable):     A callable that takes the decorated
                                    function as an argument. This cleans up the
                                    function after execution.

        Returns:
            callable:               Does the actual decoration.
    """

    def with_setup_state_wrapper(a_callable, setup=setup, teardown=teardown):
        """
            Mutates the callable s.t. it has globals for setup and teardown.

            Args:
                setup(callable):        A callable that takes the decorated
                                        function as an argument. This sets up
                                        the function before execution. Simply
                                        forwarded from before.

                teardown(callable):     A callable that takes the decorated
                                        function as an argument. This cleans up
                                        the function after execution. Simply
                                        forwarded from before.

            Returns:
                callable:               The original callable with setup and
                                        teardown globals.
        """

        stage_dict = collections.OrderedDict([
            ("setup", setup),
            ("teardown", teardown)
        ])
        stage_orderer = [(lambda a, b: (a, b)), (lambda a, b: (b, a))]
        stage_itr = zip(reversed(stage_dict.items()), stage_orderer)

        for (each_stage_name, each_new_stage), each_stage_orderer in stage_itr:
            each_old_stage = getattr(a_callable, each_stage_name, None)
            if each_new_stage:
                each_new_stage = functools.partial(each_new_stage, a_callable)

            chained_stages = None
            if each_old_stage and each_new_stage:
                first_stage, second_stage = each_stage_orderer(
                    each_old_stage, each_new_stage
                )

                def stages(first_stage=first_stage,
                           second_stage=second_stage):
                    first_stage()
                    second_stage()

                chained_stages = stages
            elif each_old_stage:
                chained_stages = each_old_stage
            elif each_new_stage:
                def a_stage(each_new_stage=each_new_stage):
                    each_new_stage()

                chained_stages = a_stage
            else:
                continue

            setattr(a_callable, each_stage_name, chained_stages)

        return(a_callable)

    return(with_setup_state_wrapper)



"""
The package ``imp`` contains algorithms useful for image processing.

===============================================================================
Overview
===============================================================================
Algorithms in the ``imp`` package include filters/convolutions, registration,
and segmentation algorithms. Together these algorithms are composed to provide
workflows for analyzing image data start to finish.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 14, 2014 20:37:08 EDT$"

__all__ = ["filters", "registration", "renorm", "segment"]

from nanshe.imp import filters
from nanshe.imp import registration
from nanshe.imp import renorm
from nanshe.imp import segment



"""
The module ``xtiff`` provides support for conversion from TIFF to HDF5.

===============================================================================
Overview
===============================================================================
The module ``xtiff`` implements a relatively simplistic form of conversion from
TIFF to HDF5. Preserves the description fields from the metadata if found as a
list under the attribute `descriptions`. Additionally, keeps track of the TIFF
filenames stitched together and the offsets of each TIFF file as the attributes
`filenames` and `offsets`, respectively.

.. todo::

    Currently, this only keeps one channel and works on one Z-plane. It would \
    be nice to relax these constraints and these features in the future.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 26, 2014 11:40:54 EDT$"


import collections

import numpy
import h5py

import vigra
import vigra.impex

from nanshe.util import iters, xglob, prof,\
    xnumpy, pathHelpers

try:
    import tifffile
except ImportError:
    # scikit-image is bundled with tifffile so use it.
    from skimage.external import tifffile


try:
    unicode
except NameError:
    unicode = str


# Get the logger
trace_logger = prof.getTraceLogger(__name__)



@prof.log_call(trace_logger)
def get_multipage_tiff_shape_dtype(new_tiff_filename):
    """
        Gets the info about the shape (including page number as time)
        and dtype.

        Args:
            new_tiff_filename(str):             the TIFF file to get info about

        Returns:
            (collections.OrderedDict):          an ordered dictionary with
                                                "shape" first and "dtype"
                                                (type) second.
    """

    shape_dtype_result = collections.OrderedDict(
        [("shape", None), ("dtype", None)])

    new_tiff_file_info = vigra.impex.ImageInfo(new_tiff_filename)
    new_tiff_file_number_pages = vigra.impex.numberImages(new_tiff_filename)

    new_tiff_file_shape = new_tiff_file_info.getShape()
    shape_dtype_result["shape"] = new_tiff_file_shape[
        :-1] + (new_tiff_file_number_pages,) + new_tiff_file_shape[-1:]

    shape_dtype_result["dtype"] = new_tiff_file_info.getDtype()

    return(shape_dtype_result)


@prof.log_call(trace_logger)
def get_multipage_tiff_shape_dtype_transformed(new_tiff_filename,
                                               axis_order="zyxtc",
                                               pages_to_channel=1):
    """
        Gets the info about the shape and dtype after some transformations
        have been performed.

        Args:
            new_tiff_filename(str):             the TIFF file to get info about
            axis_order(str):                    the desired axis order
                                                when reshaped

            pages_to_channel(int):              number of channels to divide
                                                from the pages

        Returns:
            (collections.OrderedDict):          an ordered dictionary with
                                                "shape" first and "dtype"
                                                (type) second.
    """

    assert (pages_to_channel > 0)
    assert (len(axis_order) == 5)
    assert all([_ in axis_order for _ in "zyxtc"])

    new_tiff_file_shape, new_tiff_file_dtype = list(get_multipage_tiff_shape_dtype(
        new_tiff_filename
    ).values())

    # Correct if the tiff is missing dims by adding singletons
    if (len(new_tiff_file_shape) == 5):
        pass
    elif (len(new_tiff_file_shape) == 4):
        new_tiff_file_shape = (1,) + new_tiff_file_shape
    else:
        raise Exception(
            "Invalid dimensionality for TIFF. Found shape to be \"" +
            repr(new_tiff_file_shape) + "\"."
        )

    # Correct if some pages are for different channels
    if (pages_to_channel != 1):
        new_tiff_file_shape = new_tiff_file_shape[:-2] + \
                              (new_tiff_file_shape[-2] / pages_to_channel,
                               pages_to_channel * new_tiff_file_shape[-1],)

    # Correct the axis order
    if (axis_order != "zyxtc"):
        vigra_ordering = dict(iters.reverse_each_element(enumerate("zyxtc")))

        new_tiff_file_shape_transposed = []
        for each_axis_label in axis_order:
            new_tiff_file_shape_transposed.append(
                new_tiff_file_shape[vigra_ordering[each_axis_label]]
            )

        new_tiff_file_shape = tuple(new_tiff_file_shape_transposed)

    shape_dtype_result = collections.OrderedDict(
        [("shape", None), ("dtype", None)])

    shape_dtype_result["shape"] = new_tiff_file_shape
    shape_dtype_result["dtype"] = new_tiff_file_dtype

    return(shape_dtype_result)


@prof.log_call(trace_logger)
def get_standard_tiff_array(new_tiff_filename,
                            axis_order="tzyxc",
                            pages_to_channel=1,
                            memmap=False):
    """
        Reads a tiff file and returns a standard 5D array.

        Args:
            new_tiff_filename(str):             the TIFF file to read in

            axis_order(int):                    how to order the axes (by
                                                default returns "tzyxc").

            pages_to_channel(int):              if channels are not normally
                                                stored in the channel variable,
                                                but are stored as pages (or as
                                                a mixture), then this will
                                                split neighboring pages into
                                                separate channels. (by default
                                                is 1 so changes nothing)

            memmap(bool):                       allows one to load the array
                                                using a memory mapped file as
                                                opposed to reading it directly.
                                                (by default is False)

        Returns:
            (numpy.ndarray or numpy.memmap):    an array with the axis order
                                                specified.
    """

    assert (pages_to_channel > 0)

    with tifffile.TiffFile(new_tiff_filename) as new_tiff_file:
        new_tiff_array = new_tiff_file.asarray(memmap=memmap)

    # Add a singleton channel if none is present.
    if new_tiff_array.ndim == 3:
        new_tiff_array = new_tiff_array[None]

    # Fit the old VIGRA style array. (may try to remove in the future)
    new_tiff_array = new_tiff_array.transpose(
        tuple(iters.irange(new_tiff_array.ndim - 1, 1, -1)) + (1, 0)
    )

    # Check to make sure the dimensions are ok
    if (new_tiff_array.ndim == 5):
        pass
    elif (new_tiff_array.ndim == 4):
        # Has no z. So, add this.
        new_tiff_array = xnumpy.add_singleton_axis_beginning(new_tiff_array)
    else:
        raise Exception(
            "Invalid dimensionality for TIFF. Found shape to be \"" +
            repr(new_tiff_array.shape) + "\"."
        )

    # Some people use pages to hold time and channel data. So, we need to
    # restructure it. However, if they are properly structuring their TIFF
    # file, then they shouldn't incur a penalty.
    if pages_to_channel > 1:
        new_tiff_array = new_tiff_array.reshape(
            new_tiff_array.shape[:-2] +
            (new_tiff_array.shape[-2] / pages_to_channel,
             pages_to_channel * new_tiff_array.shape[-1],)
        )

    new_tiff_array = xnumpy.tagging_reorder_array(
        new_tiff_array,
        from_axis_order="zyxtc",
        to_axis_order=axis_order,
        to_copy=True
    )

    return(new_tiff_array)


@prof.log_call(trace_logger)
def get_standard_tiff_data(new_tiff_filename,
                           axis_order="tzyxc",
                           pages_to_channel=1,
                           memmap=False):
    """
        Reads a tiff file and returns a standard 5D array and the metadata.

        Args:
            new_tiff_filename(str):             the TIFF file to read in

            axis_order(int):                    how to order the axes (by
                                                default returns "tzyxc").

            pages_to_channel(int):              if channels are not normally
                                                stored in the channel variable,
                                                but are stored as pages (or as
                                                a mixture), then this will
                                                split neighboring pages into
                                                separate channels. (by default
                                                is 1 so changes nothing)

            memmap(bool):                       allows one to load the array
                                                using a memory mapped file as
                                                opposed to reading it directly.
                                                (by default is False)

        Returns:
            (ndarray/memmap, ndarray):          an array with the axis order
                                                specified and description
                                                metadata.
    """

    assert (pages_to_channel > 0)

    new_tiff_description = []
    with tifffile.TiffFile(new_tiff_filename) as new_tiff_file:
        new_tiff_array = new_tiff_file.asarray(memmap=memmap)

        for i in iters.irange(
                0,
                len(new_tiff_file),
                pages_to_channel
        ):
            new_tiff_description.append([])
            for j in iters.irange(pages_to_channel):
                each_page = new_tiff_file[i + j]
                each_metadata = each_page.tags
                each_desc = u""

                try:
                    each_desc = unicode(
                        each_metadata["image_description"].value
                    )
                except KeyError:
                    pass

                new_tiff_description[-1].append(
                    each_desc
                )

    new_tiff_description = numpy.array(new_tiff_description)


    # Add a singleton channel if none is present.
    if new_tiff_array.ndim == 3:
        new_tiff_array = new_tiff_array[None]

    # Fit the old VIGRA style array. (may try to remove in the future)
    new_tiff_array = new_tiff_array.transpose(
        tuple(iters.irange(new_tiff_array.ndim - 1, 1, -1)) + (1, 0)
    )

    # Check to make sure the dimensions are ok
    if (new_tiff_array.ndim == 5):
        pass
    elif (new_tiff_array.ndim == 4):
        # Has no z. So, add this.
        new_tiff_array = xnumpy.add_singleton_axis_beginning(new_tiff_array)
    else:
        raise Exception(
            "Invalid dimensionality for TIFF. Found shape to be \"" +
            repr(new_tiff_array.shape) + "\"."
        )

    # Some people use pages to hold time and channel data. So, we need to
    # restructure it. However, if they are properly structuring their TIFF
    # file, then they shouldn't incur a penalty.
    if pages_to_channel > 1:
        new_tiff_array = new_tiff_array.reshape(
            new_tiff_array.shape[:-2] +
            (new_tiff_array.shape[-2] / pages_to_channel,
             pages_to_channel * new_tiff_array.shape[-1],)
        )

    new_tiff_array = xnumpy.tagging_reorder_array(
        new_tiff_array,
        from_axis_order="zyxtc",
        to_axis_order=axis_order,
        to_copy=True
    )

    # Currently is `tc` order so convert it to the expected order.
    if axis_order.index("t") > axis_order.index("c"):
        new_tiff_description = new_tiff_description.T.copy()

    return(new_tiff_array, new_tiff_description)


@prof.log_call(trace_logger)
def convert_tiffs(new_tiff_filenames,
                  new_hdf5_pathname,
                  axis=0,
                  channel=0,
                  z_index=0,
                  pages_to_channel=1,
                  memmap=False):
    """
        Convert a stack of tiffs to an HDF5 file.

        Args:
            new_tiff_filenames(list or str):    takes a str for a single file
                                                or a list of strs for
                                                filenames to combine (allows
                                                regex).

            new_hdf5_pathname(str):             the HDF5 file and location to
                                                store the dataset.

            axis(int):                          which axis to concatenate
                                                along.

            channel(int):                       which channel to select for the
                                                HDF5 (can only keep one).

            z_index(int):                       which z value to take (the
                                                algorithm is not setup for 3D
                                                data yet)

            pages_to_channel(int):              if channels are not normally
                                                stored in the channel variable,
                                                but are stored as pages, then
                                                this will split neighboring
                                                pages into separate channels.

            memmap(bool):                       allows one to load the array
                                                using a memory mapped file as
                                                opposed to reading it directly.
                                                (by default is False)
    """

    assert (pages_to_channel > 0)

    # Get the axes that do not change
    static_axes = numpy.array(list(iters.xrange_with_skip(
        3, to_skip=axis
    )))

    # if it is only a single str, make it a singleton list
    if isinstance(new_tiff_filenames, (bytes, unicode)):
        new_tiff_filenames = [new_tiff_filenames]

    # Expand any regex in path names
    new_tiff_filenames = xglob.expand_pathname_list(*new_tiff_filenames)

    # Extract the offset and descriptions for storage.
    new_hdf5_dataset_filenames = list()
    new_hdf5_dataset_offsets = list()

    # Determine the shape and dtype to use for the dataset (so that everything
    # will fit).
    new_hdf5_dataset_shape = numpy.zeros((3,), dtype=int)
    new_hdf5_dataset_dtype = bool
    for each_new_tiff_filename in new_tiff_filenames:
        # Add each filename.
        new_hdf5_dataset_filenames.append(unicode(each_new_tiff_filename))

        # Get all of the offsets.
        new_hdf5_dataset_offsets.append(new_hdf5_dataset_shape[axis])

        # Get the shape and type of each frame.
        each_new_tiff_file_shape, each_new_tiff_file_dtype = list(get_multipage_tiff_shape_dtype_transformed(
            each_new_tiff_filename,
            axis_order="cztyx",
            pages_to_channel=pages_to_channel
        ).values())
        each_new_tiff_file_shape = each_new_tiff_file_shape[2:]

        # Find the increase on the merge axis. Find the largest shape for the
        # rest.
        each_new_tiff_file_shape = numpy.array(each_new_tiff_file_shape)
        new_hdf5_dataset_shape[axis] += each_new_tiff_file_shape[axis]
        new_hdf5_dataset_shape[static_axes] = numpy.array(
            [
                new_hdf5_dataset_shape[static_axes],
                each_new_tiff_file_shape[static_axes]
            ]
        ).max(axis=0)

        # Finds the best type that everything can be cast to without loss of
        # precision.
        if not numpy.can_cast(each_new_tiff_file_dtype, new_hdf5_dataset_dtype):
            if numpy.can_cast(new_hdf5_dataset_dtype, each_new_tiff_file_dtype):
                new_hdf5_dataset_dtype = each_new_tiff_file_dtype
            else:
                raise Exception(
                    "Cannot find safe conversion between" +
                    " new_hdf5_dataset_dtype = " +
                    repr(new_hdf5_dataset_dtype) +
                    " and each_new_tiff_file_dtype = " +
                    repr(each_new_tiff_file_dtype) + "."
                )

    # Convert to arrays.
    new_hdf5_dataset_filenames = numpy.array(new_hdf5_dataset_filenames)
    new_hdf5_dataset_offsets = numpy.array(new_hdf5_dataset_offsets)

    # Convert to standard forms
    new_hdf5_dataset_shape = tuple(new_hdf5_dataset_shape)
    new_hdf5_dataset_dtype = numpy.dtype(new_hdf5_dataset_dtype)

    # Get all the needed locations for the HDF5 file and dataset
    new_hdf5_path_components = pathHelpers.PathComponents(new_hdf5_pathname)
    new_hdf5_filename = new_hdf5_path_components.externalPath
    new_hdf5_groupname = new_hdf5_path_components.internalDirectory
    new_hdf5_dataset_name = new_hdf5_path_components.internalPath

    # Dump all datasets to the file
    with h5py.File(new_hdf5_filename, "a") as new_hdf5_file:
        new_hdf5_group = new_hdf5_file.require_group(new_hdf5_groupname)
        new_hdf5_dataset = new_hdf5_group.create_dataset(
            new_hdf5_dataset_name,
            new_hdf5_dataset_shape,
            new_hdf5_dataset_dtype,
            chunks=True
        )
        new_hdf5_dataset.attrs.create(
            "filenames",
            new_hdf5_dataset_filenames,
            shape=new_hdf5_dataset_filenames.shape,
            dtype=h5py.special_dtype(vlen=unicode)
        )
        new_hdf5_dataset.attrs["offsets"] = new_hdf5_dataset_offsets
        # Workaround required due to this issue
        # ( https://github.com/h5py/h5py/issues/289 ).
        new_hdf5_descriptions_dataset = new_hdf5_group.create_dataset(
            "_".join([new_hdf5_dataset_name, "descriptions"]),
            shape=new_hdf5_dataset_shape[0:1],
            dtype=h5py.special_dtype(vlen=unicode)
        )
        new_hdf5_dataset.attrs["descriptions"] = (
            new_hdf5_descriptions_dataset.file.filename +
            new_hdf5_descriptions_dataset.name
        )

        new_hdf5_dataset_axis_pos = 0
        for each_new_tiff_filename in new_tiff_filenames:
            # Log the filename in case something goes wrong.
            trace_logger.info(
                "Now appending TIFF: \"" + str(each_new_tiff_filename) + "\""
            )

            # Read the data in the format specified.
            each_new_tiff_array, each_new_tiff_description = get_standard_tiff_data(
                each_new_tiff_filename,
                axis_order="cztyx",
                pages_to_channel=pages_to_channel,
                memmap=memmap
            )

            # Take channel and z selection
            # TODO: Could we drop the channel constraint?
            # TODO: Want to drop z constraint.
            each_new_tiff_array = each_new_tiff_array[channel, z_index]
            each_new_tiff_description = each_new_tiff_description[channel]

            # Store into the current slice and go to the next one.
            new_hdf5_dataset_axis_pos_next = new_hdf5_dataset_axis_pos + \
                                             len(each_new_tiff_array)
            new_hdf5_dataset[new_hdf5_dataset_axis_pos:new_hdf5_dataset_axis_pos_next] = each_new_tiff_array
            new_hdf5_descriptions_dataset[new_hdf5_dataset_axis_pos:new_hdf5_dataset_axis_pos_next] = each_new_tiff_description
            new_hdf5_dataset_axis_pos = new_hdf5_dataset_axis_pos_next



"""
``random_dictionary_learning_data`` supports generating synthetic data.

===============================================================================
Overview
===============================================================================
The module ``random_dictionary_learning_data`` provides a way of generating
synthetic data for testing the segmentation algorithm against. Moving forward
the useful content in here will be refactored and moved into the ``data``
module. So, depending on this module is unwise.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "Apr 17, 2014 11:27:08 EDT$"


import warnings

warnings.warn(
    "The module `random_dictionary_learning_data` is deprecated." +
    "Please consider using `data` instead." +
    "Relevant content from this module is being refactored and moved there.",
    DeprecationWarning
)

import numpy
import scipy
import scipy.ndimage

import nanshe.util.prof
import nanshe.util.iters


# Get the logger
trace_logger = nanshe.util.prof.getTraceLogger(__name__)



class MappingDiscreteUniformDistributionGenerator(object):
    """
        Given a bunch of arguments. This will create a random element generator
        that returns one or many.
    """

    def __init__(self, *args):
        """
            Builds a random element generator.

            Args:
                *args:  Anything that needs to be drawn from.
        """

        self.args = args

    def __call__(self, size=1):
        """
            Draws a certain number of elements with equal likelihood and
            returns them in a list.

            Args:
                size(int):       Number of elements to draw (one if
                                 unspecified).

            Returns:
                results(list):   a list of arguments drawn (None if no
                                 arguments)
        """

        results = []

        if len(self.args) > 0:
            indices = numpy.random.randint(0, len(self.args), size)

            results = [self.args[_] for _ in indices]
        else:
            results = [None for _ in nanshe.util.iters.irange(size)]

        return(results)


class NumpyRandomArrayDiscreteUniformDistributionGenerator(object):
    """
        Creates a random numpy array (with type bool) generator that will set
        a certain number of random positions in the array to True (like
        Bernoulli random distribution)
    """

    def __init__(self, shape):
        """
            Creates a random numpy array generator that returns arrays of a
            certain shape.

            Args:
                shape(tuple):   A tuple permissible for setting the size of a
                                numpy array.
        """

        self.shape = shape

    def __call__(self, size=1):
        """
            Generates a numpy array.

            Args:
                size(int):                Number of Trues to have in the numpy
                                          array.

            Returns:
                results(numpy.ndarray):   a boolean numpy array with a fixed
                                          number of randomly placed Trues
        """

        # A completely empty numpy array
        results = numpy.zeros(self.shape, dtype=bool)

        # Gets a set of random indices that need to be non-zero
        indices = tuple([
            numpy.random.randint(0, each_dim, size) for each_dim in self.shape
        ])

        # Makes them non-zero
        results[indices] = True

        return(results)


class MappingDiscreteGeometricDistributionGenerator(object):
    """
        A random generator of groups. Each group has a size that is
        geometrically distributed. However, the individuals chosen for the
        group are all equally likely.
    """

    def __init__(self, *args):
        """
            Sets the arguments for use to compose the groups.

            Args:
                *args:  Any variety of useful items for drawing.
        """

        self.args = args

    def __call__(self, p, size=1):
        """
            Generates a number of groups equal to size with each group size
            being distributed geometrically by p.

            Args:
                p(float)         the probability of success for a geometric
                                 distribution (starts with 1 so has mean 1/p).

                size(int)        the number of groups to make

            Returns:
                results(list):   a list of groups of arguments drawn (None if
                                 no arguments)
        """

        # Get a uniform distribution over the elements to fill each group.
        uni_gen = MappingDiscreteUniformDistributionGenerator(*self.args)

        # Draw the sizes for each group
        group_sizes = numpy.random.geometric(p, size)

        # Using the sizes draw element to fill groups up to the right size
        results = [
            uni_gen(group_sizes[i]) for i in nanshe.util.iters.irange(size)
        ]

        return(results)





class DictionaryLearningRandomDataSample(object):
    """
        Essentially a struct with its values set at runtime by
        DictionaryLearningRandomDataGenerator calls.
    """
    def __init__(self):
        """
            Default constructor just to establish values.
        """

        self.points = None
        self.centroid_activation_frames = None
        self.noiseless_frames = None
        self.frames = None


class DictionaryLearningRandomDataGenerator(object):
    """
        A Random Generator that build pseudo-data similar in nature to that
        which the ADINA algorithm is run.
    """

    def __init__(self,
                 frame_shape,
                 num_objects,
                 num_groups,
                 num_frames,
                 mean_group_size,
                 object_spread,
                 object_max_intensity,
                 object_min_intensity,
                 background_noise_intensity):
        """
            Builds a DictionaryLearningRandomDataGenerator for draws.

            Args:
                frame_shape(tuple)                  a tuple of ints for
                                                    constructing a numpy array

                num_objects(int)                    the number of objects that
                                                    can possible be active
                                                    (i.e. neurons present
                                                    whether active or not)

                num_groups(int)                     number of groups of objects
                                                    that will be active (i.e.
                                                    number of groups of neurons
                                                    seen to be active)

                num_frames(int)                     number of frames for any
                                                    group to be active in the
                                                    pseudo-video

                mean_group_size(float)              average group size (average
                                                    for a geometric
                                                    distribution)

                object_spread(float)                how big an object is on
                                                    average

                object_max_intensity(float)         the highest intensity
                                                    possible

                object_min_intensity(float)         the lowest intensity
                                                    possible

                background_noise_intensity(float)   how much noise there is in
                                                    the background.
        """

        self.frame_shape = frame_shape
        self.num_objects = num_objects
        self.num_groups = num_groups
        self.num_frames = num_frames
        self.mean_group_size = mean_group_size
        self.object_spread = object_spread
        self.object_max_intensity = object_max_intensity
        self.object_min_intensity = object_min_intensity
        self.background_noise_intensity = background_noise_intensity

        self.object_intensity_range = self.object_max_intensity - \
                                      self.object_min_intensity

    def __call__(self, num_runs=1, seed=None):
        """
            Constructs a series of pseudo-videos.

            Args:
                num_runs(int):          number of pseudo-videos to generate
                seed(int):              uses the seed for numpy.random.seed if
                                        provided.

            Returns:
                results(list):          a list of
                                        DictionaryLearningRandomDataSample
                                        instances with relevant data from
                                        generation included.
        """

        # Use the seed provided.
        numpy.random.seed(seed)

        # A list of DictionaryLearningRandomDataSample instances
        results = []

        for i in nanshe.util.iters.irange(num_runs):
            # Where the result will be stored
            each_result = DictionaryLearningRandomDataSample()

            # Generates a numpy array that has a shape of self.frame_shape with
            # a fixed number of randomly selected (equally likely) non-zero
            # entries
            each_result.points = NumpyRandomArrayDiscreteUniformDistributionGenerator(
                self.frame_shape)(self.num_objects).astype(float)

            # Creates a point generator that selects from the non-zero points
            # generated for activation to create groups
            # as an index array (tuple of 1D numpy.ndarrays)
            selected_points = each_result.points.nonzero()
            # convert to a single numpy.ndarrays
            selected_points = numpy.array(selected_points)
            # simpler, lightweight way of doing zip(*selected_points)
            selected_points = selected_points.T
            selected_points = selected_points.tolist()
            point_groups_gen = MappingDiscreteGeometricDistributionGenerator(
                *selected_points
            )

            # Using a mean group size and the number of groups creates point
            # groups (these should in someway relate to the basis images)
            point_groups = point_groups_gen(
                1.0 / float(self.mean_group_size), self.num_groups)

            # Will store the essential frames that indicate which points will
            # be active in each frame
            each_result.centroid_activation_frames = []
            for each_point_group in point_groups:
                # Get an index array
                each_point_group_index_array = nanshe.util.iters.list_indices_to_index_array(
                    each_point_group
                )

                # Create an empty activation frame
                each_centroid_activation_frame = numpy.zeros(self.frame_shape)

                # Set the active points to be randomly distributed
                each_centroid_activation_frame_points_shape = each_centroid_activation_frame[each_point_group_index_array].shape

                # Set the active points to be randomly distributed
                each_centroid_activation_frame[each_point_group_index_array] = numpy.random.random(
                    each_centroid_activation_frame_points_shape
                )

                # Rescale the active points
                each_centroid_activation_frame[each_point_group_index_array] *= self.object_intensity_range

                # Translate the active points
                each_centroid_activation_frame[each_point_group_index_array] += self.object_min_intensity

                # add to the stack of centroid activations
                each_result.centroid_activation_frames.append(
                    each_centroid_activation_frame
                )

            # convert to numpy array
            each_result.centroid_activation_frames = numpy.array(
                each_result.centroid_activation_frames
            )

            # Holds the frames without noise
            each_result.noiseless_frames = []

            # Takes each centroid activation frame and creates objects that dim
            # over time
            for each_centroid_activation_frame in each_result.centroid_activation_frames:
                # Determines how much to spread each active point
                # (self.object_spread is like the average spread)
                sigma = 2 * self.object_spread * numpy.random.random()
                for each_frame_num in nanshe.util.iters.irange(self.num_frames):
                    # Determines a linear rescaling of each image (where they
                    # slowly become dimmer)
                    rescale = float(
                        self.num_frames - each_frame_num
                    ) / float(self.num_frames)
                    # Convolves each frame to generate a frame with objects
                    # (uses the same spread for each simply dims over time)
                    each_matrix_convolved = scipy.ndimage.filters.gaussian_filter(
                        rescale * each_centroid_activation_frame, sigma
                    )
                    # Adds to the stack of frames
                    each_result.noiseless_frames.append(each_matrix_convolved)

            # Converts the form of the noiseless frames
            each_result.noiseless_frames = numpy.array(
                each_result.noiseless_frames
            )

            # Creates frames that contain some background noise from a normal
            # distribution
            each_result.frames = each_result.noiseless_frames.copy()
            each_result.frames += numpy.random.normal(
                scale=self.background_noise_intensity,
                size=each_result.frames.shape
            )

            # Append to our list of results
            results.append(each_result)


        return(results)



"""
The package ``filter`` provides a variety of transformations for image data.

===============================================================================
Overview
===============================================================================
Provides a variety of different filters for data analysis. This includes
different types of convolutions or other related transformations. Operations on
masks are supported. Computations and removal of noise are also included.
Finally, support for performing wavelet transforms is provided.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 31, 2015 22:29:31 EDT$"

__all__ = ["noise", "masks", "wavelet"]

from nanshe.imp.filters import masks
from nanshe.imp.filters import noise
from nanshe.imp.filters import wavelet



"""
The module ``search`` provides glob paths to search for content in a HDF5 file.

===============================================================================
Overview
===============================================================================
The module implements a strategy similar to Python's |glob|_ module for HDF5
files. In short, it uses regex patterns to match as many possible paths as it
can.

.. |glob| replace:: ``glob``
.. _glob: http://docs.python.org/2/library/glob.html

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 18, 2014 20:06:44 EDT$"


import re
import collections
import itertools

import h5py

from nanshe.util import iters


# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def get_matching_paths(a_filehandle, a_path_pattern):
    """
        Looks for existing paths that match the full provide pattern path.
        Returns a list of matches for the given file handle.

        Args:
            a_filehandle(h5py.File):        an HDF5 file.
            a_path_pattern(str):            an internal path (with patterns for
                                            each group) for the HDF5 file.

        Returns:
            (list):                         a list of matching paths.
    """

    current_pattern_group_matches = []

    if (isinstance(a_filehandle, h5py.Group) and a_path_pattern):
        current_group = a_filehandle

        a_path_pattern = a_path_pattern.strip("/")

        to_split = a_path_pattern.find("/")
        if to_split != -1:
            current_path = a_path_pattern[:to_split]
            next_path = a_path_pattern[1 + to_split:]
        else:
            current_path, next_path = a_path_pattern, ""

        current_pattern_group_regex = re.compile("/" + current_path + "/")

        for each_group in current_group:
            if current_pattern_group_regex.match("/" + each_group + "/") is not None:
                next_group = current_group[each_group]

                next_pattern_group_matches = get_matching_paths(
                    next_group, next_path
                )

                for each_next_pattern_group_match in next_pattern_group_matches:
                    current_pattern_group_matches.append(
                        "/" + each_group + each_next_pattern_group_match
                    )
    else:
        current_pattern_group_matches = [""]

    return(current_pattern_group_matches)


@prof.log_call(trace_logger)
def get_matching_paths_groups(a_filehandle, a_path_pattern):
    """
        Looks for parts of the path pattern and tries to match them in order.
        Returns a list of matches that can be combined to yield acceptable
        matches for the given file handle.

        Note:
            This works best when a tree structure is created systematically in
            HDF5. Then, this will recreate what the tree structure could and
            may contain.

        Args:
            a_filehandle(h5py.File):        an HDF5 file.
            a_path_pattern(str):            an internal path (with patterns for
                                            each group) for the HDF5 file.

        Returns:
            (list):                         a list of matching paths.
    """

    def get_matching_paths_groups_recursive(a_filehandle, a_path_pattern):
        current_pattern_group_matches = []

        if (isinstance(a_filehandle, h5py.Group) and a_path_pattern):
            current_pattern_group_matches.append(collections.OrderedDict())

            current_group = a_filehandle

            a_path_pattern = a_path_pattern.strip("\b").strip("/")

            to_split = a_path_pattern.find("/")
            if to_split != -1:
                current_path = a_path_pattern[:to_split]
                next_path = a_path_pattern[1 + to_split:]
            else:
                current_path, next_path = a_path_pattern, ""

            current_pattern_group_regex = re.compile("/" + current_path + "/")

            for each_group in current_group:
                if current_pattern_group_regex.match("/" + each_group + "/") is not None:
                    next_group = current_group[each_group]

                    next_pattern_group_matches = get_matching_paths_groups_recursive(
                        next_group, next_path
                    )

                    current_pattern_group_matches[0][each_group] = None

                    while (len(current_pattern_group_matches) - 1) < len(next_pattern_group_matches):
                        current_pattern_group_matches.append(
                            collections.OrderedDict()
                        )

                    for i, each_next_pattern_group_matches in enumerate(
                            next_pattern_group_matches, start=1
                    ):
                        for each_next_pattern_group_match in each_next_pattern_group_matches:
                            current_pattern_group_matches[i][each_next_pattern_group_match] = None
        else:
            current_pattern_group_matches = []

        return(current_pattern_group_matches)

    groups = get_matching_paths_groups_recursive(a_filehandle, a_path_pattern)

    new_groups = []
    for i in iters.irange(len(groups)):
        new_groups.append(list(groups[i]))

    groups = new_groups

    return(groups)


@prof.log_call(trace_logger)
def get_matching_grouped_paths(a_filehandle, a_path_pattern):
    """
        Looks for existing paths that match the full provide pattern path.
        Returns a list of matches as keys and whether they are found in the
        HDF5 file or not.

        Args:
            a_filehandle(h5py.File):        an HDF5 file.
            a_path_pattern(str):            an internal path (with patterns for
                                            each group) for the HDF5 file.

        Returns:
            (list):                         an ordered dictionary with possible
                                            paths that fit the pattern and
                                            whether they are found.
    """

    paths_found = collections.OrderedDict()

    for each_path_components in itertools.product(
            *get_matching_paths_groups(a_filehandle, a_path_pattern)
    ):
        each_path = "/" + "/".join([_ for _ in each_path_components])

        paths_found[each_path] = None

    paths_found = list(paths_found.keys())

    return(paths_found)


@prof.log_call(trace_logger)
def get_matching_grouped_paths_found(a_filehandle, a_path_pattern):
    """
        Looks for existing paths that match the full provide pattern path.
        Returns a list of matches as keys and whether they are found in the
        HDF5 file or not.

        Args:
            a_filehandle(h5py.File):        an HDF5 file.
            a_path_pattern(str):            an internal path (with patterns for
                                            each group) for the HDF5 file.

        Returns:
            (collections.OrderedDict):      an ordered dictionary with possible
                                            paths that fit the pattern and
                                            whether they are found.
    """

    paths_found = collections.OrderedDict()

    for each_path_components in itertools.product(
            *get_matching_paths_groups(a_filehandle, a_path_pattern)
    ):
        each_path = "/" + "/".join([_ for _ in each_path_components])

        paths_found[each_path] = (each_path in a_filehandle)

    return(paths_found)
